# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""DataLad pytest plugin for extensions.

This plugin provides shared pytest configuration for DataLad and its extensions.
It registers custom markers, filterwarnings, and other pytest configuration
that extensions can inherit automatically.

For DataLad extensions, simply install DataLad and the plugin will be
automatically activated via the pytest11 entry point.

The fixtures for testing DataLad itself are in datalad.conftest, which is
loaded as a regular conftest.py file when testing DataLad, but this plugin
provides the configuration that extensions can inherit.
"""

from __future__ import annotations

import re
from pathlib import Path


def pytest_configure(config):
    """Register datalad custom markers and pytest configuration.

    This allows datalad extensions to inherit markers and configuration
    without duplicating them in their own tox.ini/pytest.ini files.
    """
    # Register all custom markers used by datalad and extensions
    markers = [
        "fail_slow: marks tests that are known to fail slowly",
        "githubci_osx: marks tests for GitHub CI on macOS",
        "githubci_win: marks tests for GitHub CI on Windows",
        "integration: marks integration tests",
        "known_failure: marks tests with known failures",
        "known_failure_githubci_osx: marks tests with known failures on GitHub CI macOS",
        "known_failure_githubci_win: marks tests with known failures on GitHub CI Windows",
        "known_failure_osx: marks tests with known failures on macOS",
        "known_failure_windows: marks tests with known failures on Windows",
        "network: marks tests requiring network access",
        "osx: marks macOS-specific tests",
        "probe_known_failure: marks tests to probe if known failure is resolved",
        "serve_path_via_http: marks tests that serve paths via HTTP",
        "skip_if_adjusted_branch: skip test on adjusted branches",
        "skip_if_no_network: skip test when network is unavailable",
        "skip_if_no_psutil: skip test when psutil is not installed",
        "skip_if_on_windows: skip test on Windows",
        "skip_if_root: skip test when running as root",
        "skip_known_failure: skip tests with known failures",
        "skip_nomultiplex_ssh: skip when multiplex SSH unavailable",
        "skip_ssh: skip SSH-related tests",
        "skip_wo_symlink_capability: skip without symlink capability",
        "slow: marks slow tests (>10 seconds)",
        "turtle: marks very slow tests (>2 minutes)",
        "usecase: marks tests from user-reported use cases",
        "windows: marks Windows-specific tests",
        "with_config: marks tests with configuration",
        "with_fake_cookies_db: marks tests with fake cookies database",
        "with_memory_keyring: marks tests with memory keyring",
        "with_sameas_remotes: marks tests with sameas remotes",
        "with_testrepos: marks tests with test repositories",
        "without_http_proxy: marks tests without HTTP proxy",
        "ai_generated: marks tests generated by AI assistants",
    ]
    for marker in markers:
        config.addinivalue_line("markers", marker)

    # Register python_files pattern to ensure assertion rewriting for utils_pytest.py
    # This was previously in tox.ini as: python_files = test_*.py *_test.py utils_pytest.py
    # The first two patterns are pytest defaults, we only need to add utils_pytest.py
    config.addinivalue_line("python_files", "utils_pytest.py")

    # Register filter warnings
    # These can be overridden in extension-specific tox.ini/pytest.ini if needed
    filterwarnings = [
        "error::DeprecationWarning:^datalad",
        # TODO: https://github.com/datalad/datalad/issues/7435
        "ignore:pkg_resources is deprecated:DeprecationWarning:",
        "error:.*yield tests:pytest.PytestCollectionWarning",
        "ignore:distutils Version classes are deprecated:DeprecationWarning",
        # workaround for https://github.com/datalad/datalad/issues/6307
        "ignore:The distutils package is deprecated",
        # sent fix upstream: https://github.com/HTTPretty/HTTPretty/pull/9
        "ignore:datetime.datetime.utcnow\\(\\) is deprecated:DeprecationWarning:httpretty",
    ]
    for warning in filterwarnings:
        config.addinivalue_line("filterwarnings", warning)


def pytest_ignore_collect(collection_path: Path) -> bool:
    """Customize which files pytest collects.

    Skip old nose code and handle doctest collection carefully.
    """
    # Skip old nose code and the tests for it:
    # Note, that this is not only about executing tests but also importing those
    # files to begin with.
    if collection_path.name == "test_tests_utils.py":
        return True
    if collection_path.parts[-3:] == ("datalad", "tests", "utils.py"):
        return True
    # When pytest is told to run doctests, by default it will import every
    # source file in its search, but a number of datalad source file have
    # undesirable side effects when imported.  This hook should ensure that
    # only `test_*.py` files, `utils_pytest.py` files, and `*.py` files
    # containing doctests are imported during test collection.
    if collection_path.name.startswith("test_") or collection_path.is_dir():
        return False
    if collection_path.name == "utils_pytest.py":
        return False  # Allow utils_pytest.py to be collected
    if collection_path.suffix != ".py":
        return True
    return not any(
        re.match(r"^\s*>>>", ln) for ln in collection_path.read_text("utf-8").splitlines()
    )



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil; coding: utf-8  -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##g
"""Variety of helpers to deal with AWS S3

Use as a script to generate test buckets via e.g.

    python -m datalad.support.s3 generate test1_dirs
"""

__docformat__ = 'restructuredtext'

import logging
import mimetypes
import re
from pathlib import PurePath
from urllib.request import (
    Request,
    urlopen,
)

import boto3
from botocore.exceptions import ClientError

import datalad.log  # Just to have lgr setup happen this one used a script
from datalad.support.network import URL

lgr = logging.getLogger('datalad.s3')

# TODO: should become a config option and managed along with the rest
S3_ADMIN_CREDENTIAL = "datalad-datalad-admin-s3"
S3_TEST_CREDENTIAL = "datalad-datalad-test-s3"


def _get_s3_resource(credname=None):
    """Creates a boto3 s3 resource

    If credential name is given, DataLad credentials are retrieved or
    entered; otherwise boto3 credential discovery mechanism is used
    (~/.aws/config or env vars). Resources are a higher-level
    abstraction than clients, and seem more fitting for use in this
    module.
    """
    if credname is not None:
        from datalad.downloaders.credentials import AWS_S3
        credential = AWS_S3(credname, None)
        if not credential.is_known:
            credential.enter_new()
        creds = credential()
        session = boto3.session.Session(
            aws_access_key_id=creds["key_id"],
            aws_secret_access_key=creds["secret_id"],
        )
    else:
        session = boto3.session.Session()
    return session.resource("s3")


class VersionedFilesPool(object):
    """Just a helper which would help to create versioned files in the bucket"""
    def __init__(self, bucket):
        self._versions = {}
        self._bucket = bucket

    @property
    def bucket(self):
        return self._bucket

    def __call__(self, filename, prefix='', load=None):
        self._versions[filename] = version = self._versions.get(filename, 0) + 1
        version_str = f"version{version}"
        mtype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'

        if load is None:
            load = prefix
            if PurePath(filename).suffix == 'html':
                load += f'<html><body>{version_str}</body></html>'
            else:
                load += version_str

        object = self.bucket.Object(key=filename)
        object.put(
            Body=load.encode(),
            ContentType=mtype,
        )
        return object

    def reset_version(self, filename):
        self._versions[filename] = 0


def prune_and_delete_bucket(bucket):
    """Deletes all the content and then deletes bucket

    Should be used with care -- no confirmation requested
    """
    bucket.object_versions.delete()
    bucket.delete()
    lgr.info("Bucket %s was removed", bucket.name)


def set_bucket_public_access_policy(bucket):
    # we need to enable permissions for making content available
    bucket.Policy().put(Policy="""{
      "Version":"2012-10-17",
      "Statement":[{
          "Sid":"AddPerm",
          "Effect":"Allow",
          "Principal": "*",
          "Action":["s3:GetObject", "s3:GetObjectVersion", "s3:GetObjectTorrent", "s3:GetObjectVersionTorrent"],
          "Resource":["arn:aws:s3:::%s/*"]
        }
      ]
    }""" % bucket.name)


def gen_test_bucket(bucket_name):
    s3 = _get_s3_resource(S3_ADMIN_CREDENTIAL)
    region = s3.meta.client.meta.region_name
    bucket = s3.Bucket(bucket_name)
    # assure we have none
    exists = True
    try:
        s3.meta.client.head_bucket(Bucket=bucket_name)
    except ClientError as e:
        error_code = e.response["Error"]["Code"]
        if error_code == "404":
            exists = False
        else:
            # likely 403
            raise e
    if exists:
        lgr.info("Deleting existing bucket %s", bucket.name)
        prune_and_delete_bucket(bucket)
    # by default, bucket is created in us-east, leading to constraint exception
    # if user has different (endpoint) region in config - read & use the latter
    bucket.create(CreateBucketConfiguration={"LocationConstraint": region})
    return bucket


def _gen_bucket_test0(bucket_name="datalad-test0", versioned=True):

    bucket = gen_test_bucket(bucket_name)

    # Enable web access to that bucket to everyone
    bucket.Website().put(
        WebsiteConfiguration={"IndexDocument": {"Suffix": "index.html"}}
    )
    set_bucket_public_access_policy(bucket)

    files = VersionedFilesPool(bucket)

    files("1version-nonversioned1.txt")
    files("2versions-nonversioned1.txt")

    if versioned:
        # make bucket versioned AFTER we uploaded one file already
        bucket.Versioning().enable()

    files("2versions-nonversioned1.txt")
    files("2versions-nonversioned1.txt_sameprefix")
    for v in range(3):
        files("3versions-allversioned.txt")
    files("3versions-allversioned.txt_sameprefix")  # to test possible problems

    # File which was created and then removed
    #bucket.delete_key(files("1version-removed.txt"))
    files("1version-removed.txt").delete()

    # File which was created/removed/recreated (with new content)
    files("2versions-removed-recreated.txt").delete()
    files("2versions-removed-recreated.txt")
    files("2versions-removed-recreated.txt_sameprefix")

    # File which was created/removed/recreated (with new content)
    f = "1version-removed-recreated.txt"
    files(f).delete()
    files.reset_version(f)
    files(f)
    lgr.info("Bucket %s was generated and populated", bucket_name)

    return bucket


def gen_bucket_test0_versioned():
    return _gen_bucket_test0('datalad-test0-versioned', versioned=True)


def gen_bucket_test0_nonversioned():
    return _gen_bucket_test0('datalad-test0-nonversioned', versioned=False)


def _gen_versioned_bucket(bucket_name):
    """Create a versioned S3 test bucket with web access and public policy."""
    bucket = gen_test_bucket(bucket_name)
    bucket.Versioning().enable()
    bucket.Website().put(
        WebsiteConfiguration={"IndexDocument": {"Suffix": "index.html"}}
    )
    set_bucket_public_access_policy(bucket)
    return bucket, VersionedFilesPool(bucket)


def gen_bucket_test1_dirs():
    bucket, files = _gen_versioned_bucket('datalad-test1-dirs-versioned')

    files("d1", load="")  # creating an empty file
    # then we would like to remove that d1 as a file and make a directory out of it
    files("d1/file1.txt")
    # and then delete it and place it back
    files("d1", load="smth")


def gen_bucket_test2_obscurenames_versioned():
    # in principle bucket name could also contain ., but boto doesn't digest it
    # well
    bucket, files = _gen_versioned_bucket('datalad-test2-obscurenames-versioned')

    # http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingMetadata.html
    files("f 1", load="")
    files("f [1][2]")
    # Need to grow up for this .... TODO
    #files(u"юникод")
    #files(u"юни/код")
    # all fancy ones at once
    files("f!-_.*'( )")
    # the super-fancy which aren't guaranteed to be good idea (as well as [] above)
    files("f &$=@:+,?;")


def gen_bucket_test1_manydirs():
    # to test crawling with flexible subdatasets making decisions
    bucket, files = _gen_versioned_bucket('datalad-test1-manydirs-versioned')

    files("d1", load="")  # creating an empty file
    # then we would like to remove that d1 as a file and make a directory out of it
    files("d1/file1.txt")
    files("d1/sd1/file1.txt")
    files("d1/sd2/file3.txt", load="a")
    files("d1/sd2/ssd1/file4.txt")
    files("d2/file1.txt")
    files("d2/sd1/file1.txt")
    files("d2/sd1/ssd/sssd/file1.txt")


def add_version_to_url(url, version, replace=False):
    """Add a version ID to `url`.

    Parameters
    ----------
    url : datalad.support.network.URL
        A URL.
    version : str
        The value of 'versionId='.
    replace : boolean, optional
        If a versionID is already present in `url`, replace it.

    Returns
    -------
    A versioned URL (str)
    """
    version_id = "versionId={}".format(version)
    if not url.query:
        query = version_id
    else:
        ver_match = re.match("(?P<pre>.*&)?"
                             "(?P<vers>versionId=[^&]+)"
                             "(?P<post>&.*)?",
                             url.query)
        if ver_match:
            if replace:
                query = "".join([ver_match.group("pre") or "",
                                 version_id,
                                 ver_match.group("post") or ""])
            else:
                query = url.query
        else:
            query = url.query + "&" + version_id
    return URL(**dict(url.fields, query=query)).as_str()


def get_versioned_url(url, guarantee_versioned=False, return_all=False, verify=False,
                      s3client=None, update=False):
    """Given a url return a versioned URL

    Originally targeting AWS S3 buckets with versioning enabled

    Parameters
    ----------
    url : string
    guarantee_versioned : bool, optional
      Would fail if buckets is determined to have no versioning enabled.
      It will not fail if we fail to determine if bucket is versioned or
      not
    return_all: bool, optional
      If True, would return a list with URLs for all the versions of this
      file, sorted chronologically with latest first (when possible, e.g.
      for S3).  Remove markers get ignored
    verify: bool, optional
      Verify that URL is accessible. As discovered some versioned keys might
      be denied access to
    s3client: botocore.client.S3, optional
      A boto3 client instance that will be used to interact with AWS; if None,
      a new one will be created.
    update : bool, optional
      If the URL already contains a version ID, update it to the latest version
      ID.  This option has no effect if return_all is true.

    Returns
    -------
    string or list of string
    """
    url_rec = URL(url)

    s3_bucket, fpath = None, url_rec.path.lstrip('/')

    was_versioned = False
    all_versions = []


    # hostname regex match allowing optional region code
    # bucket-name.s3.region-code.amazonaws.com
    match_virtual_hosted_style = re.match(
        r"^(.+)(\.s3)(?:[.-][a-z0-9-]+){0,1}(\.amazonaws\.com)$", url_rec.hostname
    )
    # s3.region-code.amazonaws.com/bucket-name/key-name
    match_path_style = re.match(
        r"^s3(?:\.[a-z0-9-]+){0,1}(\.amazonaws\.com)$", url_rec.hostname
    )

    if match_virtual_hosted_style is not None:
        s3_bucket = match_virtual_hosted_style.group(1)
    elif match_path_style is not None:
        if url_rec.scheme not in ('http', 'https'):
            raise ValueError("Do not know how to handle %s scheme" % url_rec.scheme)
        # url is s3.amazonaws.com/bucket/PATH
        s3_bucket, fpath = fpath.split('/', 1)
    elif url_rec.scheme == 's3':
        s3_bucket = url_rec.hostname  # must be
        if url_rec.query and 'versionId=' in url_rec.query:
            was_versioned = True
            all_versions.append(url)
        else:
            # and for now implement magical conversion to URL
            # TODO: wouldn't work if needs special permissions etc
            # actually for now
            raise NotImplementedError

    if s3_bucket:
        # TODO: cache
        if s3client is None:
            # we need to reuse our providers
            from ..downloaders.providers import Providers
            providers = Providers.from_config_files()
            s3url = "s3://%s/" % s3_bucket
            s3provider = providers.get_provider(s3url)
            authenticator = s3provider.authenticator
            if not authenticator:
                # we will use the default one
                from ..downloaders.s3 import S3Authenticator
                authenticator = S3Authenticator()
            if authenticator.client is not None:
                # we have established connection before, so let's just reuse
                s3client = authenticator.client
            else:
                s3client = authenticator.authenticate(s3_bucket, s3provider.credential)

        supports_versioning = True  # assume that it does
        try:
            # Status can be "Enabled" | "Suspended", or missing altogether
            response = s3client.get_bucket_versioning(Bucket=s3_bucket)
            supports_versioning = response.get("Status") == "Enabled"
        except ClientError as e:
            # might be forbidden, i.e. "403 Forbidden" so we try then anyways
            supports_versioning = 'maybe'

        if supports_versioning:
            response = s3client.list_object_versions(
                Bucket=s3_bucket,
                Prefix=fpath,
            )

            all_keys = response.get("Versions", [])
            # Filter and sort them so the newest one on top
            all_keys = [
                x
                for x in sorted(
                    all_keys,
                    key=lambda x: (x["LastModified"], x["IsLatest"]),
                    reverse=True,
                )
                if ((x["Key"] == fpath))  # match exact name, not just prefix
            ]

            # our current assumptions
            assert all_keys[0]["IsLatest"]

            # boto compatibility note: boto3 response should separate
            # Versions & DeleteMarkers, no action needed

            for key in all_keys:
                url_versioned = add_version_to_url(
                    url_rec, key["VersionId"], replace=update and not return_all)

                all_versions.append(url_versioned)
                if verify:
                    # it would throw HTTPError exception if not accessible
                    _ = urlopen(Request(url_versioned))
                was_versioned = True
                if not return_all:
                    break

    if guarantee_versioned and not was_versioned:
        raise RuntimeError("Could not version %s" % url)

    if not all_versions:
        # we didn't get a chance
        all_versions = [url_rec.as_str()]

    if return_all:
        return all_versions
    else:
        return all_versions[0]


if __name__ == '__main__':
    import sys
    lgr.setLevel(logging.INFO)
    # TODO: proper cmdline
    if len(sys.argv) > 1 and sys.argv[1] == "generate":
        if len(sys.argv) < 3:
            raise ValueError("Say 'all' to regenerate all, or give a generators name")
        name = sys.argv[2]
        if name.lower() == 'all':
            gb = [k for k in locals().keys() if k.startswith('gen_bucket')]
            for func_name in gb:
                locals()[func_name]()
        else:
            locals()['gen_bucket_%s' % name]()
    else:
        print("nothing to do")



"""DataLad pytest configuration for local testing.

This conftest.py provides fixtures and configuration for testing DataLad itself.

The pytest hooks (pytest_configure, pytest_ignore_collect) are defined in
datalad.pytest_plugin and are imported here so they're active when testing
DataLad locally. For DataLad extensions, those hooks are loaded via the
pytest11 entry point.
"""

import logging
import os
import re
from contextlib import ExitStack
from pathlib import Path
from unittest.mock import patch

import pytest

from datalad.support.exceptions import CapturedException
from datalad.utils import (
    get_encoding_info,
    get_envvars_info,
    get_home_envvars,
)

from . import (
    cfg,
    ssh_manager,
)
from .log import lgr
# Import pytest hooks from plugin so they're active for local testing
# This is safe - when testing datalad itself, this conftest is loaded
# as a regular conftest (not via entry point), so no double registration
from .pytest_plugin import (  # noqa: F401
    pytest_configure,
    pytest_ignore_collect,
)

_test_states = {}

# handle to an HTTP server instance that is used as part of the tests
test_http_server = None

@pytest.fixture(autouse=True, scope="session")
def _fix_pwd_for_changedir():
    """Ensure PWD env var matches os.getcwd().

    When tox uses changedir, subprocess CWD differs from inherited PWD.
    Without this fix, getpwd() permanently switches to os.getcwd() mode,
    breaking symlink path preservation and CWD recovery.
    """
    from datalad import utils
    os.environ['PWD'] = os.getcwd()
    utils._pwd_mode = None


@pytest.fixture(autouse=True, scope="session")
def setup_package():
    import tempfile

    from datalad import consts
    from datalad.support.annexrepo import AnnexRepo
    from datalad.support.cookies import cookies_db
    from datalad.support.external_versions import external_versions
    from datalad.tests import _TEMP_PATHS_GENERATED
    from datalad.tests.utils_pytest import (
        DEFAULT_BRANCH,
        DEFAULT_REMOTE,
        OBSCURE_FILENAME,
        HTTPPath,
        rmtemp,
    )
    from datalad.ui import ui
    from datalad.utils import (
        make_tempfile,
        on_osx,
    )

    if on_osx:
        # enforce honoring TMPDIR (see gh-5307)
        tempfile.tempdir = os.environ.get('TMPDIR', tempfile.gettempdir())

    # Use unittest's patch instead of pytest.MonkeyPatch for compatibility with
    # old pytests
    with ExitStack() as m:
        m.enter_context(patch.object(consts, "DATASETS_TOPURL", 'https://datasets-tests.datalad.org/'))
        m.enter_context(patch.dict(os.environ, {'DATALAD_DATASETS_TOPURL': consts.DATASETS_TOPURL}))

        m.enter_context(
            patch.dict(
                os.environ,
                {
                    "GIT_CONFIG_PARAMETERS":
                    "'init.defaultBranch={}' 'clone.defaultRemoteName={}'"
                    .format(DEFAULT_BRANCH, DEFAULT_REMOTE)
                }
            )
        )
        cred_cfg = cfg.obtain('datalad.tests.credentials')
        if cred_cfg == 'plaintext':
            m.enter_context(
                patch.dict(
                    os.environ,
                    {
                        'PYTHON_KEYRING_BACKEND':
                            'keyrings.alt.file.PlaintextKeyring'
                    }
                )
            )
        elif cred_cfg == 'system':
            pass
        else:
            raise ValueError(cred_cfg)

        def prep_tmphome():
            # re core.askPass:
            # Don't let git ask for credentials in CI runs. Note, that this variable
            # technically is not a flag, but an executable (which is why name and value
            # are a bit confusing here - we just want a no-op basically). The environment
            # variable GIT_ASKPASS overwrites this, but neither env var nor this config
            # are supported by git-credential on all systems and git versions (most recent
            # ones should work either way, though). Hence use both across CI builds.
            gitconfig = """\
[user]
        name = DataLad Tester
        email = test@example.com
[core]
	askPass =
[datalad "log"]
        exc = 1
[annex "security"]
	# from annex 6.20180626 file:/// and http://localhost access isn't
	# allowed by default
	allowed-url-schemes = http https file
	allowed-http-addresses = all
[protocol "file"]
    # since git 2.38.1 cannot by default use local clones for submodules
    # https://github.blog/2022-10-18-git-security-vulnerabilities-announced/#cve-2022-39253
    allow = always
""" + os.environ.get('DATALAD_TESTS_GITCONFIG', '').replace('\\n', os.linesep)
            # TODO: split into a function + context manager
            with make_tempfile(mkdir=True) as new_home:
                pass
            # register for clean-up on exit
            _TEMP_PATHS_GENERATED.append(new_home)

            # populate default config
            new_home = Path(new_home)
            new_home.mkdir(parents=True, exist_ok=True)
            cfg_file = new_home / '.gitconfig'
            cfg_file.write_text(gitconfig)
            return new_home, cfg_file

        if external_versions['cmd:git'] < "2.32":
            # To overcome pybuild overriding HOME but us possibly wanting our
            # own HOME where we pre-setup git for testing (name, email)
            if 'GIT_HOME' in os.environ:
                m.enter_context(patch.dict(os.environ, {'HOME': os.environ['GIT_HOME']}))
            else:
                # we setup our own new HOME, the BEST and HUGE one
                new_home, _ = prep_tmphome()
                m.enter_context(patch.dict(os.environ, get_home_envvars(new_home)))
        else:
            _, cfg_file = prep_tmphome()
            m.enter_context(patch.dict(os.environ, {'GIT_CONFIG_GLOBAL': str(cfg_file)}))

        # Re-load ConfigManager, since otherwise it won't consider global config
        # from new $HOME (see gh-4153
        cfg.reload(force=True)

        # datalad.locations.sockets has likely changed. Discard any cached values.
        ssh_manager._socket_dir = None

        # To overcome pybuild by default defining http{,s}_proxy we would need
        # to define them to e.g. empty value so it wouldn't bother touching them.
        # But then haskell libraries do not digest empty value nicely, so we just
        # pop them out from the environment
        for ev in ('http_proxy', 'https_proxy'):
            if ev in os.environ and not (os.environ[ev]):
                lgr.debug("Removing %s from the environment since it is empty", ev)
                os.environ.pop(ev)

        # Prevent interactive credential entry (note "true" is the command to run)
        # See also the core.askPass setting above
        m.enter_context(patch.dict(os.environ, {'GIT_ASKPASS': 'true'}))

        # Set to non-interactive UI
        _test_states['ui_backend'] = ui.backend
        # obtain() since that one consults for the default value
        ui.set_backend(cfg.obtain('datalad.tests.ui.backend'))

        # in order to avoid having to fiddle with rather uncommon
        # file:// URLs in the tests, have a standard HTTP server
        # that serves an 'httpserve' directory in the test HOME
        # the URL will be available from datalad.test_http_server.url

        global test_http_server
        # Start the server only if not running already
        # Relevant: we have test_misc.py:test_test which runs datalad.test but
        # not doing teardown, so the original server might never get stopped
        if test_http_server is None:
            serve_path = tempfile.mkdtemp(
                dir=cfg.get("datalad.tests.temp.dir"),
                prefix='httpserve',
            )
            test_http_server = HTTPPath(serve_path)
            test_http_server.start()
            _TEMP_PATHS_GENERATED.append(serve_path)

        yield

        lgr.debug("Printing versioning information collected so far")
        # Query for version of datalad, so it is included in ev.dumps below - useful while
        # testing extensions where version of datalad might differ in the environment.
        external_versions['datalad']
        print(external_versions.dumps(query=True))
        try:
            print("Obscure filename: str=%s repr=%r"
                    % (OBSCURE_FILENAME.encode('utf-8'), OBSCURE_FILENAME))
        except UnicodeEncodeError as exc:
            ce = CapturedException(exc)
            print("Obscure filename failed to print: %s" % ce)
        def print_dict(d):
            return " ".join("%s=%r" % v for v in d.items())
        print("Encodings: %s" % print_dict(get_encoding_info()))
        print("Environment: %s" % print_dict(get_envvars_info()))

        if os.environ.get('DATALAD_TESTS_NOTEARDOWN'):
            return

        ui.set_backend(_test_states['ui_backend'])

        if test_http_server:
            test_http_server.stop()
            test_http_server = None
        else:
            lgr.debug("For some reason global http_server was not set/running, thus not stopping")

        if len(_TEMP_PATHS_GENERATED):
            msg = "Removing %d dirs/files: %s" % (len(_TEMP_PATHS_GENERATED), ', '.join(_TEMP_PATHS_GENERATED))
        else:
            msg = "Nothing to remove"
        lgr.debug("Teardown tests. " + msg)
        for path in _TEMP_PATHS_GENERATED:
            rmtemp(str(path), ignore_errors=True)

    # Re-establish correct global config after changing $HOME.
    # Might be superfluous, since after teardown datalad.cfg shouldn't be
    # needed. However, maintaining a consistent state seems a good thing
    # either way.
    cfg.reload(force=True)

    ssh_manager._socket_dir = None

    cookies_db.close()


@pytest.fixture(autouse=True)
def capture_logs(caplog, monkeypatch):
    DATALAD_LOG_LEVEL = os.environ.get('DATALAD_LOG_LEVEL', None)
    if DATALAD_LOG_LEVEL is None:
        # very very silent.  Tests introspecting logs should use
        # swallow_logs(new_level=...)
        caplog.set_level(100, lgr.name)
        # And we should also set it within environ so underlying commands also
        # stay silent
        monkeypatch.setenv('DATALAD_LOG_LEVEL', '100')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Base DataLad command execution runner
"""
from __future__ import annotations

import logging
from os import PathLike
from queue import Queue
from typing import (
    IO,
    cast,
)

from .coreprotocols import NoCapture
from .exception import CommandError
from .nonasyncrunner import (
    ThreadedRunner,
    _ResultGenerator,
)
from .protocol import (
    GeneratorMixIn,
    WitlessProtocol,
)

lgr = logging.getLogger('datalad.runner.runner')


class WitlessRunner(object):
    """Minimal Runner with support for online command output processing

    It aims to be as simple as possible, providing only essential
    functionality.
    """
    __slots__ = ['cwd', 'env']

    def __init__(self,
                 cwd: str | PathLike | None = None,
                 env: dict | None = None
                 ):
        """
        Parameters
        ----------
        cwd : str or path-like, optional
          If given, commands are executed with this path as PWD,
          the PWD of the parent process is used otherwise.
        env : dict, optional
          Environment to be used for command execution. If `None` is provided,
          the current process environment is inherited by the sub-process. If
          a mapping is provided, it will constitute the complete environment of
          the sub-process, i.e.  no values from the environment of the current
          process will be inherited. If `env` and `cwd` are given, 'PWD' in the
          environment is set to the string-value of `cwd`.
        """
        self.env = env
        self.cwd = cwd

    def _get_adjusted_env(self,
                          env: dict | None = None,
                          cwd: str | PathLike | None = None,
                          copy: bool = True
                          ) -> dict | None:
        """Return an adjusted execution environment

        This method adjusts the environment provided in `env` to
        reflect the configuration of the runner. It returns
        an altered copy or an altered original, if `copy` is
        `False`.

        Parameters
        ----------
        env
          The environment that should be adjusted

        cwd: str | PathLike | None (default: None)
          If not None, the content of this variable will be
          put into the environment variable 'PWD'.

        copy: bool (default: True)
          if True, the returned environment will be a
          copy of `env`. Else the passed in environment
          is modified. Note: if `env` is not `None` and
          `cwd` is `None` and `copy` is `True`, the
          returned dictionary is still a copy
        """
        env = env.copy() if env is not None and copy is True else env
        if cwd is not None and env is not None:
            # If an environment and 'cwd' is provided, ensure the 'PWD' in the
            # environment is set to the value of 'cwd'.
            env['PWD'] = str(cwd)
        return env

    def run(self,
            cmd: list | str,
            protocol: type[WitlessProtocol] | None = None,
            stdin: bytes | IO | Queue | None = None,
            cwd: PathLike | str | None = None,
            env: dict | None = None,
            timeout: float | None = None,
            exception_on_error: bool = True,
            **kwargs) -> dict | _ResultGenerator:
        """Execute a command and communicate with it.

        Parameters
        ----------
        cmd : list or str
          Sequence of program arguments. Passing a single string causes
          execution via the platform shell.
        protocol : WitlessProtocol, optional
          Protocol class handling interaction with the running process
          (e.g. output capture). A number of pre-crafted classes are
          provided (e.g `KillOutput`, `NoCapture`, `GitProgress`).
          If the protocol has the GeneratorMixIn-mixin, the run-method
          will return an iterator and can therefore be used in a for-clause.
        stdin : file-like, bytes, Queue, or None
          If stdin is a file-like, it will be directly used as stdin for the
          subprocess. The caller is responsible for writing to it and closing it.
          If stdin is a bytes, it will be fed to stdin of the subprocess.
          If all data is written, stdin will be closed.
          If stdin is a Queue, all elements (bytes) put into the Queue will
          be passed to stdin until None is read from the queue. If None is read,
          stdin of the subprocess is closed.
        cwd : str or path-like, optional
          If given, commands are executed with this path as PWD,
          the PWD of the parent process is used otherwise. Overrides
          any `cwd` given to the constructor.
        env : dict, optional
          Environment to be used for command execution. If given, it will
          completely replace any environment provided to theconstructor. If
          `cwd` is given, 'PWD' in the environment is set to its value.
          This must be a complete environment definition, no values
          from the current environment will be inherited. Overrides
          any `env` given to the constructor.
        timeout: float, optional
          None or the seconds after which a timeout callback is
          invoked, if no progress was made in communicating with
          the sub-process, or if waiting for the subprocess exit
          took more than the specified time. See the protocol and
          `ThreadedRunner` descriptions for a more detailed discussion
          on timeouts.
        exception_on_error : bool, optional
          This argument is first interpreted if the protocol is a subclass
          of `GeneratorMixIn`. If it is `True` (default), a
          `CommandErrorException` is raised by the generator if the
          sub process exited with a return code not equal to zero. If the
          parameter is `False`, no exception is raised. In both cases the
          return code can be read from the attribute `return_code` of
          the generator. Then this argument interpreted within this function
          to not raise `CommandError` if value is False in case of non-0 exit.
        kwargs :
          Passed to the Protocol class constructor.

        Returns
        -------
          dict | _ResultGenerator

            If the protocol is not a subclass of `GeneratorMixIn`, the
            result of protocol._prepare_result will be returned.

            If the protocol is a subclass of `GeneratorMixIn`, a Generator, i.e.
            a `_ResultGenerator`, will be returned. This allows to use this
            method in constructs like:

                for protocol_output in runner.run():
                    ...

            Where the iterator yields whatever protocol.pipe_data_received
            sends into the generator.
            If all output was yielded and the process has terminated, the
            generator will raise StopIteration(return_code), where
            return_code is the return code of the process. The return code
            of the process will also be stored in the "return_code"-attribute
            of the runner. So you could write:

               gen = runner.run()
               for file_descriptor, data in gen:
                   ...

               # get the return code of the process
               result = gen.return_code

        Raises
        ------
        CommandError
          On execution failure (non-zero exit code) this exception is
          raised which provides the command (cmd), stdout, stderr,
          exit code (status), and a message identifying the failed
          command, as properties.
        FileNotFoundError
          When a given executable does not exist.
        """
        if protocol is None:
            # by default let all subprocess stream pass through
            protocol = NoCapture

        applied_cwd = cwd or self.cwd
        applied_env = self._get_adjusted_env(
            env=env or self.env,
            cwd=applied_cwd,
        )

        lgr.debug(
            'Run %r (protocol_class=%s) (cwd=%s)',
            cmd,
            protocol.__name__,
            applied_cwd
        )

        threaded_runner = ThreadedRunner(
            cmd=cmd,
            protocol_class=protocol,
            stdin=stdin,
            protocol_kwargs=kwargs,
            timeout=timeout,
            exception_on_error=exception_on_error,
            cwd=applied_cwd,
            env=applied_env
        )

        results_or_iterator = threaded_runner.run()
        if issubclass(protocol, GeneratorMixIn):
            return results_or_iterator

        results = cast(dict, results_or_iterator)
        # log before any exception is raised
        lgr.debug("Finished %r with status %s", cmd, results['code'])

        # make it such that we always blow if a protocol did not report
        # a return code at all or it was non-0 and we were not asked ignore
        # errors
        return_code = results.get('code', None)
        if return_code is None or (return_code and exception_on_error):
            # the runner has a better idea, doc string warns Protocol
            # implementations not to return these
            results.pop('cmd', None)
            results.pop('cwd', None)
            raise CommandError(
                # whatever the results were, we carry them forward
                cmd=cmd,
                cwd=applied_cwd,
                **results,
            )
        # Denoise result, the return code must be zero at this point.
        results.pop('code', None)
        return results



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Class the starts a subprocess and keeps it around to communicate with it
via stdin. For each instruction send over stdin, a response is read and
returned. The response structure is determined by "output_proc"

"""
from __future__ import annotations

import logging
import os
import queue
import sys
import warnings
from datetime import datetime
from operator import attrgetter
from queue import Queue
from subprocess import TimeoutExpired
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)
from weakref import (
    ReferenceType,
    WeakValueDictionary,
    ref,
)

from datalad import cfg
# start of legacy import block
# to avoid breakage of code written before datalad.runner
from datalad.runner.coreprotocols import (
    KillOutput,
    NoCapture,
    StdErrCapture,
    StdOutCapture,
    StdOutErrCapture,
)
from datalad.runner.gitrunner import (
    GIT_SSH_COMMAND,
    GitRunnerBase,
    GitWitlessRunner,
)
from datalad.runner.nonasyncrunner import (
    STDERR_FILENO,
    STDOUT_FILENO,
    _ResultGenerator,
    run_command,
)
from datalad.runner.protocol import (
    GeneratorMixIn,
    WitlessProtocol,
)
from datalad.runner.runner import WitlessRunner
from datalad.runner.utils import LineSplitter
from datalad.support.exceptions import CommandError
from datalad.utils import (
    auto_repr,
    ensure_unicode,
)

# end of legacy import block



__docformat__ = "restructuredtext"


_cfg_var = "datalad.runtime.stalled-external"
_cfg_val = cfg.obtain(_cfg_var)


class BatchedCommandError(CommandError):
    def __init__(self,
                 cmd="",
                 last_processed_request="",
                 msg="",
                 code=None,
                 stdout="",
                 stderr="",
                 cwd=None,
                 **kwargs):
        """
        This exception extends a CommandError that is raised by the command,
        that is executed by `BatchedCommand`. It extends the `CommandError` by
        `last_processed_request`. This attribute contains the last request, i.e.
        argument to `BatchedCommand.__call__()`, that was successfully
        processed, i.e. for which a result was received from the command (that
        does not imply that the result was positive).

        :param last_processed_request: the last request for which a response was
            received from the underlying command. This could be used to restart
            an interrupted process.

        For all other arguments see `CommandError`.
        """
        CommandError.__init__(
            self,
            cmd=cmd,
            msg=msg,
            code=code,
            stdout=stdout,
            stderr=stderr,
            cwd=cwd,
            **kwargs
        )
        self.last_processed_request = last_processed_request


lgr = logging.getLogger('datalad.cmd')

# TODO unused?
# In python3 to split byte stream on newline, it must be bytes
linesep_bytes = os.linesep.encode()

# TODO unused?
_TEMP_std = sys.stdout, sys.stderr

# TODO unused?
# To be used in the temp file name to distinguish the ones we create
# in Runner so we take care about their removal, in contrast to those
# which might be created outside and passed into Runner
_MAGICAL_OUTPUT_MARKER = "_runneroutput_"


def readline_rstripped(stdout):
    warnings.warn("the function `readline_rstripped()` is deprecated "
                  "and will be removed in a future release",
                  DeprecationWarning)
    return _readline_rstripped(stdout)


def _readline_rstripped(stdout):
    """Internal helper for BatchedCommand"""
    return stdout.readline().rstrip()


class BatchedCommandProtocol(GeneratorMixIn, StdOutErrCapture):
    def __init__(self,
                 batched_command: "BatchedCommand",
                 done_future: Any = None,
                 encoding: Optional[str] = None,
                 output_proc: Optional[Callable] = None,
                 ):
        GeneratorMixIn.__init__(self)
        StdOutErrCapture.__init__(self, done_future, encoding)
        self.batched_command = batched_command
        self.output_proc = output_proc
        self.line_splitter = LineSplitter()

    def pipe_data_received(self, fd: int, data: bytes):
        if fd == STDERR_FILENO:
            self.send_result((fd, data))
        elif fd == STDOUT_FILENO:
            for line in self.line_splitter.process(data.decode(self.encoding)):
                self.send_result((fd, line))
        else:
            raise ValueError(f"unknown file descriptor: {fd}")

    def pipe_connection_lost(self, fd: int, exc: Optional[BaseException]):
        if fd == STDOUT_FILENO:
            remaining_line = self.line_splitter.finish_processing()
            if remaining_line is not None:
                lgr.debug("unterminated line: %s", remaining_line)
                self.send_result((fd, remaining_line))

    def timeout(self, fd: Optional[int]) -> bool:
        timeout_error = self.batched_command.get_timeout_exception(fd)
        if timeout_error:
            raise timeout_error
        self.send_result(("timeout", fd))
        return False


class ReadlineEmulator:
    """
    This class implements readline() on the basis of an instance of
    BatchedCommand. Its purpose is to emulate stdout's for output_procs,
    This allows us to provide a BatchedCommand API that is identical
    to the old version, but with an implementation that is based on the
    threaded runner.
    """
    def __init__(self,
                 batched_command: "BatchedCommand"):
        self.batched_command = batched_command

    def readline(self):
        """
        Read from the stdout provider until we have a line or None (which
        indicates some error).
        """
        return self.batched_command.get_one_line()


class SafeDelCloseMixin(object):
    """A helper class to use where __del__ would call .close() which might
    fail if "too late in GC game"
    """
    def __del__(self):
        try:
            self.close()
        except (TypeError, ImportError):
            # ImportError could be raised when the interpreter is shutting down.
            if os.fdopen is None or lgr.debug is None:
                # if we are late in the game and things already gc'ed in py3,
                # it is Ok
                return
            raise


@auto_repr
class BatchedCommand(SafeDelCloseMixin):
    """
    Container for a running subprocess. Supports communication with the
    subprocess via stdin and stdout.
    """

    # Collection of active BatchedCommands as a mapping from object IDs to
    # instances
    _active_instances: WeakValueDictionary[int, BatchedCommand] = WeakValueDictionary()

    def __init__(self,
                 cmd: Union[str, Tuple, List],
                 path: Optional[str] = None,
                 output_proc: Optional[Callable] = None,
                 timeout: Optional[float] = None,
                 exception_on_timeout: bool = False,
                 ):

        command = cmd
        self.command: list = [command] if not isinstance(command, List) else command
        self.path: Optional[str] = path
        self.output_proc: Optional[Callable] = output_proc
        self.timeout: Optional[float] = timeout
        self.exception_on_timeout: bool = exception_on_timeout

        self.stderr_output = b""
        self.runner: Optional[WitlessRunner] = None
        self.encoding = None
        self.wait_timed_out = None
        self.return_code: Optional[int] = None
        self._abandon_cache = None
        self.last_request: Optional[str] = None

        self._active = 0
        self._active_last = _now()
        self.clean_inactive()
        assert id(self) not in self._active_instances
        self._active_instances[id(self)] = self

        # pure declarations
        self.stdin_queue: Queue
        self.generator: _ResultGenerator

    @classmethod
    def clean_inactive(cls):
        from . import cfg
        max_batched = cfg.obtain("datalad.runtime.max-batched")
        max_inactive_age = cfg.obtain("datalad.runtime.max-inactive-age")
        if len(cls._active_instances) > max_batched:
            active_qty = 0
            inactive = []
            for c in cls._active_instances.values():
                if c._active:
                    active_qty += 1
                else:
                    inactive.append(c)
            inactive.sort(key=attrgetter("_active_last"))
            to_close = len(cls._active_instances) - max_batched
            if to_close <= 0:
                return
            too_young = 0
            now = _now()
            for i, c in enumerate(inactive):
                if (now - c._active_last).total_seconds() <= max_inactive_age:
                    too_young = len(inactive) - i
                    break
                elif c._active:
                    active_qty += 1
                else:
                    c.close()
                    cls._active_instances.pop(id(c), None)
                    to_close -= 1
                    if to_close <= 0:
                        break
            if to_close > 0:
                lgr.debug(
                    "Too many BatchedCommands remaining after cleanup;"
                    " %d active, %d went inactive recently",
                    active_qty,
                    too_young,
                )

    def _initialize(self):

        lgr.debug("Starting new runner for %s", repr(self))
        lgr.log(5, "Command: %s", self.command)

        self.stdin_queue = queue.Queue()
        self.stderr_output = b""
        self.wait_timed_out = None
        self.return_code = None
        self.last_request = None

        self.runner = WitlessRunner(
            cwd=self.path,
            env=GitRunnerBase.get_git_environ_adjusted()
        )
        self.generator = self.runner.run(
            cmd=self.command,
            protocol=BatchedCommandProtocol,
            stdin=self.stdin_queue,
            cwd=self.path,
            # This mimics the behavior of the old implementation w.r.t
            # timeouts when waiting for the closing process
            timeout=self.timeout or 11.0,
            # Keyword arguments for the protocol
            batched_command=self,
            output_proc=self.output_proc,
        )
        self.encoding = self.generator.runner.protocol.encoding

        self._active_last = _now()

    def process_running(self) -> bool:
        if self.runner:
            if self.generator.runner.process is None:
                return False
            result = self.generator.runner.process.poll()
            if result is None:
                return True
            self.return_code = result
            self.runner = None
            if result != 0:
                raise BatchedCommandError(
                    cmd=" ".join(self.command),
                    last_processed_request=self.last_request,
                    msg=f"{type(self).__name__}: exited with {result} after "
                        f"request: {self.last_request}",
                    code=result
                ) from CommandError
        return False

    def __call__(self,
                 cmds: Union[str, Tuple, List]):
        """
        Send requests to the subprocess and return the responses. We expect one
        response per request. How the response is structured is determined by
        output_proc. If output_proc returns not-None, the responses is
        considered to be a response.

        If output_proc is not provided, we assume that a single response is
        a single line.

        If the subprocess does not exist yet it is started before the first
        command is sent.

        Parameters
        ----------
        cmds : str or tuple or list of (str or tuple)
            request for the subprocess

        Returns
        -------
        (return_type[self.output_proc] | str)
        | list[(return_type[self.output_proc] | str)]

            Responses received from process. Either a single element, or a list
            of elements, if `cmds` was a list.
            The type of the elements is `str`, if `self.output_proc` is `None`.
            If `self.output_proc` is not `None`, the result type of
            `self.output_proc` determines the type of the elements.
        """
        self._active += 1
        requests = cmds

        input_multiple = isinstance(requests, list)
        if not input_multiple:
            requests = [requests]

        responses = []
        try:
            # This code assumes that each processing request is
            # a single line and leads to a response that triggers a
            # `send_result` in the protocol.
            for request in requests:
                while True:
                    try:
                        responses.append(self.process_request(request))
                        self.last_request = request
                        break
                    except StopIteration:
                        # The process finished executing, store the last return
                        # code and restart the process.
                        lgr.debug("%s: command exited", self)
                        self.return_code = self.generator.return_code
                        self.runner = None

        except CommandError as command_error:
            # Convert CommandError into BatchedCommandError
            self.runner = None
            self.return_code = command_error.code
            raise BatchedCommandError(
                cmd=command_error.cmd,
                last_processed_request=self.last_request,
                msg=command_error.msg,
                code=command_error.code,
                stdout=command_error.stdout,
                stderr=command_error.stderr,
                cwd=command_error.cwd,
                **command_error.kwargs
            ) from command_error

        finally:
            self._active -= 1
        return responses if input_multiple else responses[0] if responses else None

    def process_request(self,
                        request: Union[Tuple, str]) -> Any | None:

        self._active += 1
        try:

            if not self.process_running():
                self._initialize()

            # Remember request and send it to subprocess
            if not isinstance(request, str):
                request = ' '.join(request)
            self.stdin_queue.put((request + "\n").encode())

            # Get the response from the generator. We only consider
            # data received on stdout as a response.
            if self.output_proc:
                # If we have an output procedure, let the output procedure
                # read stdout and decide about the nature of the response
                response = self.output_proc(ReadlineEmulator(self))
            else:
                # If there is no output procedure we assume that a response
                # is one line.
                response = self.get_one_line()
                if response is not None:
                    response = response.rstrip()
            return response

        finally:
            self._active -= 1

    def proc1(self,
              single_command: str):
        """
        Simulate the old interface. This method is used only once in
        AnnexRepo.get_metadata()
        """
        self._active += 1
        try:
            assert isinstance(single_command, str)
            return self(single_command)
        finally:
            self._active -= 1

    def get_one_line(self) -> Optional[str]:
        """
        Get a single stdout line from the generator.

        If timeout was specified, and exception_on_timeout is False,
        and if a timeout occurs, return None. Otherwise, return the
        string that was read from the generator.
        """

        # Implementation remarks:
        # 1. We know that BatchedCommandProtocol only returns complete lines on
        #    stdout, that makes this code simple.
        # 2. stderr is handled transparently within this method,
        #    by adding all stderr-content to an internal buffer.
        while True:
            source, data = self.generator.send(None)
            if source == STDERR_FILENO:
                self.stderr_output += data
            elif source == STDOUT_FILENO:
                return data
            elif source == "timeout":
                # TODO: we should restart the subprocess on timeout, otherwise
                #  we might end up with results from a previous instruction,
                #  when handling multiple instructions at once. Until this is
                #  done properly, communication timeouts are ignored in order
                #  to avoid errors.
                if data is None:
                    lgr.debug('BatchedCommand: timeout on process')
                else:
                    lgr.debug('BatchedCommand: timeout on file descriptor %d', data)
            else:
                raise ValueError(f"{self}: unknown source: {source}")

    def close(self, return_stderr=False):
        """
        Close communication and wait for process to terminate. If the "timeout"
        parameter to the constructor was not None, and if the configuration
        setting "datalad.runtime.stalled-external" is set to "abandon",
        the method will return latest after "timeout" seconds. If the subprocess
        did not exit within this time, the attribute "wait_timed_out" will
        be set to "True".

        Parameters
        ----------
        return_stderr: bool
          if set to "True", the call will return all collected stderr content
          as string. In addition, if return_stderr is True and the log level
          is 5 or lower, and the configuration setting "datalad.log.outputs"
          evaluates to "True", the content of stderr will be logged.

        Returns
        -------
        str, optional
          stderr output if return_stderr is True, None otherwise
        """

        if self.runner:

            abandon = self._get_abandon()

            # Close stdin to let the process know that we want to end
            # communication. We also close stdout and stderr to inform
            # the generator that we do not care about them anymore. This
            # will trigger process wait timeouts.
            self.generator.runner.close_stdin()

            # Process all remaining messages until the subprocess exits.
            remaining = []
            timeout = False
            try:
                for source, data in self.generator:
                    if source == STDERR_FILENO:
                        self.stderr_output += data
                    elif source == STDOUT_FILENO:
                        remaining.append(data)
                    elif source == "timeout":
                        if data is None and abandon is True:
                            timeout = True
                            break
                    else:
                        raise ValueError(f"{self}: unknown source: {source}")
                self.return_code = self.generator.return_code

            except CommandError as command_error:
                lgr.error(
                    "%s subprocess exited with %s (%s)",
                    self,
                    repr(command_error.code),
                    command_error
                )
                self.return_code = command_error.code

            if remaining:
                lgr.debug("%s: remaining content: %s", self, remaining)

            self.wait_timed_out = timeout is True
            if self.wait_timed_out:
                lgr.debug(
                    "%s: timeout while waiting for subprocess to exit", self)
                lgr.warning(
                    "Batched process (%s) "
                    "did not finish, abandoning it without killing it",
                    self.generator.runner.process.pid,
                )

        result = self.get_requested_error_output(return_stderr)
        self.runner = None
        self.stderr_output = b""
        return result

    def get_requested_error_output(self, return_stderr: bool):
        if not self.runner:
            return None

        stderr_content = ensure_unicode(self.stderr_output)
        if lgr.isEnabledFor(5):
            from . import cfg
            if cfg.getbool("datalad.log", "outputs", default=False):
                stderr_lines = stderr_content.splitlines()
                lgr.log(
                    5,
                    "stderr of %s had %d lines:",
                    self.generator.runner.process.pid if self.generator.runner.process else 'terminated',
                    len(stderr_lines))
                for line in stderr_lines:
                    lgr.log(5, "| " + line)
        if return_stderr:
            return stderr_content
        return None

    def get_timeout_exception(self,
                              fd: Optional[int]
                              ) -> Optional[TimeoutExpired]:
        """
        Get a process timeout exception if timeout exceptions should
        be generated for a process that continues longer than timeout
        seconds after self.close() was initiated.
        """
        if self.timeout is None \
                or fd is not None \
                or self.exception_on_timeout is False\
                or self._get_abandon() == "wait":
            return None
        return TimeoutExpired(
            cmd=self.command,
            timeout=self.timeout or 11.0,
            stderr=self.stderr_output)

    def _get_abandon(self):
        if self._abandon_cache is None:
            if _cfg_val not in ("wait", "abandon"):
                raise ValueError(f"Unexpected value: {_cfg_var}={_cfg_val!r}")
            self._abandon_cache = _cfg_val == "abandon"
        return self._abandon_cache


def _now():
    return datetime.now().astimezone()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from packaging.version import Version

import datalad
from datalad.support.network import (
    get_url_response_stamp,
    is_url_quoted,
)

from .utils_pytest import *


def test_is_url_quoted():
    ok_(is_url_quoted('%22%27%3ba&b&cd|'))
    ok_(not is_url_quoted('a b'))


def test_get_response_stamp():
    r = get_url_response_stamp("http://www.example.com/1.dat",
                           {'Content-length': '101',
                            'Last-modified': 'Wed, 01 May 2013 03:02:00 GMT'})
    eq_(r['size'], 101)
    eq_(r['mtime'], 1367377320)
    eq_(r['url'], "http://www.example.com/1.dat")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper functionality and overloads for paths treatment

One of the reasons is also to robustify operation with unicode filenames
"""
from __future__ import annotations

# TODO: RF and move all paths related functions from datalad.utils in here
import os
import os.path as op
# to not pollute API importing as _
from collections import defaultdict as _defaultdict
from collections.abc import (
    Iterable,
    Iterator,
)
from functools import wraps
from itertools import dropwhile
from pathlib import (
    Path,
    PurePosixPath,
)

from ..utils import (
    ensure_bytes,
    getpwd,
)


def _get_unicode_robust_version(f):

    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except UnicodeEncodeError:
            return f(ensure_bytes(*args, **kwargs))
    doc = getattr(f, '__doc__', None)
    # adjust only if __doc__ is not completely absent (None)
    if doc is not None:
        wrapped.__doc__ = doc + \
            "\n\nThis wrapper around original function would encode forcefully " \
            "to utf-8 if initial invocation fails"
    return wrapped


abspath = op.abspath
basename = op.basename
curdir = op.curdir
dirname = op.dirname
exists = _get_unicode_robust_version(op.exists)
isdir = _get_unicode_robust_version(op.isdir)
isabs = _get_unicode_robust_version(op.isabs)
join = op.join
lexists = _get_unicode_robust_version(op.lexists)
normpath = op.normpath
pardir = op.pardir
pathsep = op.pathsep
relpath = op.relpath
realpath = _get_unicode_robust_version(op.realpath)
sep = op.sep


def robust_abspath(p: str | Path) -> str:
    """A helper which would not fail if p is relative and we are in non-existing directory

    It will rely on getpwd, which would rely on $PWD env variable to report
    the path.  Desired for improved resilience during e.g. reporting as in
    https://github.com/datalad/datalad/issues/2787
    """
    try:
        return abspath(p)
    except OSError:
        if not isabs(p):
            try:
                os.getcwd()
            except Exception:
                return normpath(join(getpwd(), p))
        # if no exception raised it was not the reason, raise original
        raise


def split_ext(filename: str) -> tuple[str, str]:
    """Use git-annex's splitShortExtensions rule for splitting extensions.

    Parameters
    ----------
    filename : str

    Returns
    -------
    A tuple with (root, extension)

    Examples
    --------
    >>> from datalad.local.addurls import split_ext
    >>> split_ext("filename.py")
    ('filename', '.py')

    >>> split_ext("filename.tar.gz")
    ('filename', '.tar.gz')

    >>> split_ext("filename.above4chars.ext")
    ('filename.above4chars', '.ext')
    """
    parts = filename.split(".")
    if len(parts) == 1:
        return filename, ""

    tail = list(dropwhile(lambda x: len(x) < 5,
                          reversed(parts[1:])))

    file_parts = parts[:1] + tail[::-1]
    ext_parts = parts[1+len(tail):]
    return ".".join(file_parts), "." + ".".join(ext_parts)


def get_parent_paths(paths: list[str], parents: list[str], only_with_parents: bool = False, *, sep: str = '/') -> list[str]:
    """Given a list of children paths, return their parent paths among parents
    or their own path if there is no known parent. A path is also considered its
    own parent (haven't you watched Predestination?) ;)

    All paths should be relative, not pointing outside (not starting
    with ../), and normalized (no // or dir/../dir and alike). Only minimal
    sanity checking of values is done.  By default paths are considered to be
    POSIX. Use 'sep' kwarg to set to `os.sep` to provide OS specific handling.

    Accent is made on performance to avoid O(len(paths) * len(parents))
    runtime.  ATM should be typically less than O(len(paths) * len(log(parents)))

    Initial intended use - for a list of paths in the repository
    to provide their paths as files/submodules known to that repository, to
    overcome difference in ls-tree and ls-files, where ls-files outputs nothing
    for paths within submodules.
    It is coded, so it could later be applied even whenever there are nested
    parents, e.g. parents = ['sub', 'sub/sub'] and then the "deepest" parent
    is selected

    Parameters
    ----------
    parents: list of str
    paths: list of str
    only_with_parents: bool, optional
      If set to True, return a list of only parent paths where that path had
      a parent
    sep: str, optional
      Path separator.  By default - '/' and thus treating paths as POSIX.
      If you are processing OS-specific paths (for both `parents` and `paths`),
      specify `sep=os.sep`.

    Returns
    -------
    A list of paths (without duplicates), where some entries replaced with
    their "parents" without duplicates.  So for 'a/b' and 'a/c' with a being
    among parents, there will be a single 'a'
    """
    # Let's do an early check even though then we would skip the checks on paths
    # being relative etc
    if not parents:
        return [] if only_with_parents else paths

    # We will create a lookup for known parent lengths
    parent_set = set(parents)  # O(log(len(parents))) lookup

    # Will be used in sanity checking that we got consistently used separators, i.e.
    # not mixing non-POSIX paths and POSIX parents
    asep = {'/': '\\', '\\': '/'}[sep]

    # rely on path[:n] be quick, and len(parent_lengths) << len(parent_set)
    # when len(parent_set) is large.  We will also bail checking any parent of
    # the length if at that length path has no directory boundary ('/').
    #
    # Create mapping for each length of
    # parent path to list of parents with that length
    parent_lengths_map: dict[int, set[str]] = _defaultdict(set)
    for parent in parent_set:
        _get_parent_paths_check(parent, sep, asep)
        parent_lengths_map[len(parent)].add(parent)

    # Make it ordered in the descending order so we select the deepest/longest parent
    # and store them as sets for faster lookup.
    # Could be an ordered dict but no need
    parent_lengths = [(l, parent_lengths_map[l]) for l in sorted(parent_lengths_map, reverse=True)]

    res = []
    seen = set()

    for path in paths:  # O(len(paths)) - unavoidable but could be parallelized!
        # Sanity check -- should not be too expensive
        _get_parent_paths_check(path, sep, asep)
        for parent_length, parents_ in parent_lengths:  # O(len(parent_lengths))
            if (len(path) < parent_length) or (len(path) > parent_length and path[parent_length] != sep):
                continue  # no directory deep enough
            candidate_parent = path[:parent_length]
            if candidate_parent in parents_:  # O(log(len(parent_set))) but expected one less due to per length handling
                if candidate_parent not in seen:
                    res.append(candidate_parent)
                    seen.add(candidate_parent)
                break  # it is!
        else:  # no hits
            if not only_with_parents:
                if path not in seen:
                    res.append(path)
                    seen.add(path)

    return res


def get_filtered_paths_(paths: Iterable[str|Path], filter_paths: Iterable[str | Path],
                        *, include_within_path: bool = False) \
        -> Iterator[str]:
    """Among paths (or Path objects) select the ones within filter_paths.

    All `paths` and `filter_paths` must be relative and POSIX.

    In case of `include_with_path=True`, if a `filter_path` points to some path
    under a `path` within `paths`, that path would be returned as well, e.g.
    `path` 'submod' would be returned if there is a `filter_path` 'submod/subsub/file'.

    Complexity is O(N*log(N)), where N is the largest of the lengths of `paths`
    or `filter_paths`.

    Yields
    ------
    paths, sorted (so order is not preserved), which reside under 'filter_paths' or
    path within 'filter_paths' is under that path.
    """
    # do conversion and sanity checks, O(N)
    def _harmonize_paths(l: Iterable[str | Path]) -> list[tuple[str, ...]]:
        ps = []
        for p in l:
            pp = PurePosixPath(p)
            if pp.is_absolute():
                raise ValueError(f"Got absolute path {p}, expected relative")
            if pp.parts and pp.parts[0] == '..':
                raise ValueError(f"Path {p} leads outside")
            ps.append(pp.parts)  # store parts
        return sorted(ps)  # O(N * log(N))

    paths_parts = _harmonize_paths(paths)
    filter_paths_parts = _harmonize_paths(filter_paths)

    # we will pretty much "scroll" through sorted paths and filter_paths at the same time
    for path_parts in paths_parts:
        while filter_paths_parts:
            filter_path_parts = filter_paths_parts[0]
            l = min(len(path_parts), len(filter_path_parts))
            # if common part is "greater" in the path -- we can go to the next "filter"
            if filter_path_parts[:l] < path_parts[:l]:
                # get to the next one
                filter_paths_parts = filter_paths_parts[1:]
            else:
                break  # otherwise -- consider this one!
        else:
            # no filter path left - the other paths cannot be the selected ones
            break
        if include_within_path:
            # if one identical or subpath of another one -- their parts match in the beginning
            # and we will just reuse that 'l'
            pass
        else:
            # if all components of the filter match, for that we also add len(path_parts) check below
            l = len(filter_path_parts)
        if len(path_parts) >= l and (path_parts[:l] == filter_path_parts[:l]):
            yield '/'.join(path_parts)


def _get_parent_paths_check(path: str, sep: str, asep: str) -> None:
    """A little helper for get_parent_paths"""
    if isabs(path) or path.startswith(pardir + sep) or path.startswith(curdir + sep):
        raise ValueError("Expected relative within directory paths, got %r" % path)
    if sep+sep in path:
        raise ValueError(f"Expected normalized paths, got {path} containing '{sep+sep}'")
    if asep in path:
        raise ValueError(f"Expected paths with {sep} as separator, got {path} containing '{asep}'")



# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test the thread based runner (aka. non asyncio based runner).
"""
from __future__ import annotations

import asyncio
import os
import queue
import signal
import subprocess
import sys
import time
from collections.abc import (
    Generator,
    Iterator,
)
from itertools import count
from queue import Queue
from threading import Thread
from time import sleep
from typing import (
    Any,
    Optional,
)
from unittest.mock import (
    MagicMock,
    patch,
)

import pytest

from datalad.tests.utils_pytest import (
    assert_false,
    assert_raises,
    assert_true,
    eq_,
    known_failure_osx,
    known_failure_windows,
    with_tempfile,
)
from datalad.utils import on_windows

from .. import (
    NoCapture,
    Protocol,
    Runner,
    StdOutCapture,
    StdOutErrCapture,
)
from ..nonasyncrunner import (
    IOState,
    ThreadedRunner,
    run_command,
)
from ..protocol import GeneratorMixIn
from ..runnerthreads import (
    ReadThread,
    WriteThread,
)
from ..utils import LineSplitter
from .utils import py2cmd


# Protocol classes used for a set of generator tests later
class GenStdoutStderr(GeneratorMixIn, StdOutErrCapture):
    def __init__(self,
                 done_future: Any = None,
                 encoding: Optional[str] = None) -> None:

        StdOutErrCapture.__init__(
            self,
            done_future=done_future,
            encoding=encoding)
        GeneratorMixIn.__init__(self)

    def timeout(self, fd: Optional[int]) -> bool:
        return True


class GenNothing(GeneratorMixIn, NoCapture):
    def __init__(self,
                 done_future: Any = None,
                 encoding: Optional[str] = None) -> None:

        NoCapture.__init__(
            self,
            done_future=done_future,
            encoding=encoding)
        GeneratorMixIn.__init__(self)


class GenStdoutLines(GeneratorMixIn, StdOutCapture):
    """A generator-based protocol yielding individual subprocess' stdout lines

    This is a simple implementation that is good enough for tests, i.e. with
    controlled inpute. It will fail if data is delivered in parts to
    self.pipe_data_received that are split inside an encoded character.
    """
    def __init__(self,
                 done_future: Any = None,
                 encoding: Optional[str] = None) -> None:

        StdOutCapture.__init__(
            self,
            done_future=done_future,
            encoding=encoding)
        GeneratorMixIn.__init__(self)
        self.line_splitter = LineSplitter()

    def timeout(self, fd: Optional[int]) -> bool:
        return True

    def pipe_data_received(self, fd: int, data: bytes) -> None:
        for line in self.line_splitter.process(data.decode(self.encoding)):
            self.send_result(line)

    def pipe_connection_lost(self, fd: int, exc: Optional[BaseException]) -> None:
        remaining_line = self.line_splitter.finish_processing()
        if remaining_line is not None:
            self.send_result(remaining_line)


def test_subprocess_return_code_capture() -> None:

    class KillProtocol(Protocol):

        proc_out = True
        proc_err = True

        def __init__(self, signal_to_send: int, result_pool: dict) -> None:
            super().__init__()
            self.signal_to_send = signal_to_send
            self.result_pool = result_pool

        def connection_made(self, process: subprocess.Popen) -> None:
            super().connection_made(process)
            process.send_signal(self.signal_to_send)

        def connection_lost(self, exc: Optional[BaseException]) -> None:
            self.result_pool["connection_lost_called"] = (True, exc)

        def process_exited(self) -> None:
            self.result_pool["process_exited_called"] = True

    # windows doesn't support SIGINT but would need a Ctrl-C
    signal_to_send = signal.SIGTERM if on_windows else signal.SIGINT
    result_pool: dict[str, Any] = dict()
    result = run_command(["waitfor", "/T", "10000", "TheComputerTurnsIntoATulip"]
                         if on_windows
                         else ["sleep", "10000"],
                         KillProtocol,
                         None,
                         {
                             "signal_to_send": signal_to_send,
                             "result_pool": result_pool
                         },
                         exception_on_error=False)
    assert isinstance(result, dict)
    if not on_windows:
        # this one specifically tests the SIGINT case, which is not supported
        # on windows
        eq_(result["code"], -signal_to_send)
    assert_true(result_pool["connection_lost_called"][0])
    assert_true(result_pool["process_exited_called"])


def test_interactive_communication() -> None:

    class BidirectionalProtocol(Protocol):

        proc_out = True
        proc_err = True

        def __init__(self, result_pool: dict[str, bool]) -> None:
            super().__init__()
            self.state = 0
            self.result_pool = result_pool

        def connection_made(self, process: subprocess.Popen) -> None:
            super().connection_made(process)
            assert self.process is not None
            assert self.process.stdin is not None
            os.write(self.process.stdin.fileno(), b"1 + 1\n")

        def connection_lost(self, exc: Optional[BaseException]) -> None:
            self.result_pool["connection_lost_called"] = True

        def process_exited(self) -> None:
            self.result_pool["process_exited_called"] = True

        def pipe_data_received(self, fd: int, data: bytes) -> None:
            super().pipe_data_received(fd, data)
            assert self.process is not None
            assert self.process.stdin is not None
            if self.state == 0:
                self.state += 1
                os.write(self.process.stdin.fileno(), b"2 ** 3\n")
            if self.state == 1:
                self.state += 1
                os.write(self.process.stdin.fileno(), b"exit(0)\n")

    result_pool: dict[str, bool] = dict()
    result = run_command([sys.executable, "-i"],
                         BidirectionalProtocol,
                         stdin=subprocess.PIPE,
                         protocol_kwargs={
                            "result_pool": result_pool
                         })

    assert isinstance(result, dict)
    lines = [line.strip() for line in result["stdout"].splitlines()]
    eq_(lines, ["2", "8"])
    assert_true(result_pool["connection_lost_called"], True)
    assert_true(result_pool["process_exited_called"], True)


def test_blocking_thread_exit() -> None:
    read_queue: Queue[tuple[Any, IOState, bytes]] = queue.Queue()

    (read_descriptor, write_descriptor) = os.pipe()
    read_file = os.fdopen(read_descriptor, "rb")
    read_thread = ReadThread(
        identifier="test thread",
        user_info=read_descriptor,
        source=read_file,
        destination_queue=read_queue,
        signal_queues=[]
    )
    read_thread.start()

    os.write(write_descriptor, b"some data")
    assert_true(read_thread.is_alive())
    identifier, state, data = read_queue.get()
    eq_(data, b"some data")

    read_thread.request_exit()

    # Check the blocking part
    sleep(.3)
    assert_true(read_thread.is_alive())

    # Check actual exit, we will not get
    # "more data" when exit was requested,
    # because the thread will not attempt
    # a write
    os.write(write_descriptor, b"more data")
    read_thread.join()
    print(read_queue.queue)
    assert_true(read_queue.empty())


def test_blocking_read_exception_catching() -> None:
    read_queue: Queue[tuple[Any, IOState, Any]] = queue.Queue()

    (read_descriptor, write_descriptor) = os.pipe()
    read_file = os.fdopen(read_descriptor, "rb")
    read_thread = ReadThread(
        identifier="test thread",
        user_info=read_descriptor,
        source=read_file,
        destination_queue=read_queue,
        signal_queues=[read_queue]
    )
    read_thread.start()

    os.write(write_descriptor, b"some data")
    assert_true(read_thread.is_alive())
    identifier, state, data = read_queue.get()
    eq_(data, b"some data")
    os.close(write_descriptor)
    read_thread.join()
    identifier, state, data = read_queue.get()
    eq_(data, None)


def test_blocking_read_closing() -> None:
    # Expect that a reader thread exits when os.read throws an error.
    fake_file = MagicMock(**{"fileno.return_value": -1, "close.return_value": None})

    def fake_read(*args: Any) -> None:
        raise ValueError("test exception")

    read_queue: Queue[tuple[Any, IOState, Optional[bytes]]] = queue.Queue()
    destination_queue: Queue[tuple[Any, IOState, bytes]] = queue.Queue()
    with patch("datalad.runner.runnerthreads.os.read") as read:
        read.side_effect = fake_read

        read_thread = ReadThread(
            identifier="test thread",
            user_info=None,
            source=fake_file,
            destination_queue=destination_queue,
            signal_queues=[read_queue])

        read_thread.start()
        read_thread.join()

    identifier, state, data = read_queue.get()
    eq_(data, None)


def test_blocking_write_exception_catching() -> None:
    # Expect that a blocking writer catches exceptions and exits gracefully.

    write_queue: Queue[Optional[bytes]] = queue.Queue()
    signal_queue: Queue[tuple[Any, IOState, Optional[bytes]]] = queue.Queue()

    (read_descriptor, write_descriptor) = os.pipe()
    write_file = os.fdopen(write_descriptor, "rb")
    write_thread = WriteThread(
        identifier="test thread",
        user_info=write_descriptor,
        source_queue=write_queue,
        destination=write_file,
        signal_queues=[signal_queue]
    )
    write_thread.start()

    write_queue.put(b"some data")
    data = os.read(read_descriptor, 1024)
    eq_(data, b"some data")

    os.close(read_descriptor)
    os.close(write_descriptor)

    write_queue.put(b"more data")
    write_thread.join()
    eq_(signal_queue.get(), (write_descriptor, IOState.ok, None))


def test_blocking_writer_closing() -> None:
    # Expect that a blocking writer closes its file when `None` is sent to it.
    write_queue: Queue[Optional[bytes]] = queue.Queue()
    signal_queue: Queue[tuple[Any, IOState, Optional[bytes]]] = queue.Queue()

    (read_descriptor, write_descriptor) = os.pipe()
    write_file = os.fdopen(write_descriptor, "rb")
    write_thread = WriteThread(
        identifier="test thread",
        user_info=write_descriptor,
        source_queue=write_queue,
        destination=write_file,
        signal_queues=[signal_queue]
    )
    write_thread.start()

    write_queue.put(b"some data")
    data = os.read(read_descriptor, 1024)
    eq_(data, b"some data")

    write_queue.put(None)
    write_thread.join()
    eq_(signal_queue.get(), (write_descriptor, IOState.ok, None))


def test_blocking_writer_closing_timeout_signal() -> None:
    # Expect that writer or reader do not block forever on a full signal queue

    write_queue: Queue[Optional[bytes]] = queue.Queue()
    signal_queue: Queue[tuple[Any, IOState, Optional[bytes]]] = queue.Queue(1)
    signal_queue.put(("This is data", IOState.ok, None))

    (read_descriptor, write_descriptor) = os.pipe()
    write_file = os.fdopen(write_descriptor, "rb")
    write_thread = WriteThread(
        identifier="test thread",
        user_info=write_descriptor,
        source_queue=write_queue,
        destination=write_file,
        signal_queues=[signal_queue]
    )
    write_thread.start()

    write_queue.put(b"some data")
    data = os.read(read_descriptor, 1024)
    eq_(data, b"some data")

    write_queue.put(None)
    write_thread.join()
    eq_(signal_queue.get(), ("This is data", IOState.ok, None))


def test_blocking_writer_closing_no_signal() -> None:
    # Expect that writer or reader do not block forever on a full signal queue

    write_queue: Queue[Optional[bytes]] = queue.Queue()
    signal_queue: Queue[tuple[Any, IOState, Optional[bytes]]] = queue.Queue(1)
    signal_queue.put(("This is data", IOState.ok, None))

    (read_descriptor, write_descriptor) = os.pipe()
    write_file = os.fdopen(write_descriptor, "rb")
    write_thread = WriteThread(
        identifier="test thread",
        user_info=write_descriptor,
        source_queue=write_queue,
        destination=write_file,
        signal_queues=[signal_queue]
    )
    write_thread.start()

    write_queue.put(b"some data")
    data = os.read(read_descriptor, 1024)
    eq_(data, b"some data")

    write_queue.put(None)
    write_thread.join()


def test_inside_async() -> None:
    async def main() -> dict:
        runner = Runner()
        res = runner.run(
            (["cmd.exe", "/c"] if on_windows else []) + ["echo", "abc"],
            StdOutCapture)
        assert isinstance(res, dict)
        return res

    result = asyncio.run(main())
    eq_(result["stdout"], "abc" + os.linesep)


# Helper function for test_popen_invocation - must be at module level to be picklable
def _get_file_in_subprocess(ds_path: str, file_path: str) -> None:
    """Fetch a file in a subprocess by reconstructing the Dataset from path."""
    from datalad.distribution.dataset import Dataset
    Dataset(ds_path).get(path=file_path)


# Both Windows and OSX suffer from wrapt's object proxy insufficiency
# NotImplementedError: object proxy must define __reduce_ex__()
@known_failure_osx
@known_failure_windows
@with_tempfile(mkdir=True)
@with_tempfile
def test_popen_invocation(src_path: str = "", dest_path: str = "") -> None:
    # https://github.com/ReproNim/testkraken/issues/93
    # Note: Python 3.14+ changed default multiprocessing start method from
    # 'fork' to 'forkserver' on Linux. 'forkserver' requires pickling arguments,
    # so we pass the path string instead of a Dataset object/bound method.
    # This pattern (reconstructing Dataset from path in subprocess) is the
    # recommended way to use datalad with multiprocessing on all platforms.
    from multiprocessing import Process

    from datalad.api import clone  # type: ignore[attr-defined]
    from datalad.distribution.dataset import Dataset

    src = Dataset(src_path).create()
    (src.pathobj / "file.dat").write_bytes(b"\000")
    src.save(message="got data")

    clone(source=src_path, path=dest_path)
    fetching_data = Process(
        target=_get_file_in_subprocess,
        args=(dest_path, 'file.dat')
    )
    fetching_data.start()
    fetching_data.join(5.0)
    assert_false(fetching_data.is_alive(), "Child is stuck!")


def test_timeout() -> None:
    # Expect timeout protocol calls on long running process
    # if the specified timeout is short enough
    class TestProtocol(StdOutErrCapture):

        received_timeouts: list[tuple[int, Optional[int]]] = []

        def __init__(self) -> None:
            StdOutErrCapture.__init__(self)
            self.counter = count()

        def timeout(self, fd: Optional[int]) -> bool:
            TestProtocol.received_timeouts.append((next(self.counter), fd))
            return False

    run_command(
        ["waitfor", "/T", "1", "TheComputerTurnsIntoATulip"]
        if on_windows
        else ["sleep", "1"],
        stdin=None,
        protocol=TestProtocol,
        timeout=.1
    )
    assert_true(len(TestProtocol.received_timeouts) > 0)
    assert_true(all(map(lambda e: e[1] in (1, 2, None), TestProtocol.received_timeouts)))


def test_timeout_nothing() -> None:
    # Expect timeout protocol calls for the process on long running processes,
    # if the specified timeout is short enough.
    class TestProtocol(NoCapture):
        def __init__(self,
                     timeout_queue: list[Optional[int]]) -> None:
            NoCapture.__init__(self)
            self.timeout_queue = timeout_queue
            self.counter = count()

        def timeout(self, fd: Optional[int]) -> bool:
            self.timeout_queue.append(fd)
            return False

    stdin_queue: Queue[Optional[bytes]] = queue.Queue()
    for i in range(12):
        stdin_queue.put(b"\x00" * 1024)
    stdin_queue.put(None)

    timeout_queue: list[Optional[int]] = []
    run_command(
        py2cmd("import time; time.sleep(.4)\n"),
        stdin=stdin_queue,
        protocol=TestProtocol,
        timeout=.1,
        protocol_kwargs=dict(timeout_queue=timeout_queue)
    )
    # Ensure that we have only process timeouts and at least one
    assert_true(all(map(lambda e: e is None, timeout_queue)))
    assert_true(len(timeout_queue) > 0)


class _TimeoutProtocol(StdOutErrCapture):
    def __init__(self,
                 timeout_queue: list[tuple[int, Optional[int]]]) -> None:
        StdOutErrCapture.__init__(self)
        self.timeout_queue = timeout_queue
        self.counter = count()

    def timeout(self, fd: Optional[int]) -> bool:
        self.timeout_queue.append((next(self.counter), fd))
        return False


def _run_timeout_command() -> list[tuple[int, Optional[int]]]:
    stdin_queue: Queue[Optional[bytes]] = queue.Queue()
    for i in range(12):
        stdin_queue.put(b"\x00" * 1024)
    stdin_queue.put(None)

    timeout_queue: list[tuple[int, Optional[int]]] = []
    run_command(
        py2cmd("import time;time.sleep(.5)\n"),
        stdin=stdin_queue,
        protocol=_TimeoutProtocol,
        timeout=.1,
        protocol_kwargs=dict(timeout_queue=timeout_queue)
    )
    return timeout_queue


def test_timeout_stdout_stderr() -> None:
    # Expect timeouts on stdout and stderr
    timeout_queue = _run_timeout_command()

    # Expect at least one timeout for stdout and stderr.
    # there might be more.
    sources = (1, 2)
    assert_true(len(timeout_queue) >= len(sources))
    for source in sources:
        assert_true(any(fd == source for _, fd in timeout_queue))


def test_timeout_process() -> None:
    # Expect timeouts on stdout, stderr, and the process
    timeout_queue = _run_timeout_command()

    # Expect at least one timeout for stdout and stderr.
    sources = (1, 2)
    assert_true(len(timeout_queue) >= len(sources))
    for source in sources:
        assert_true(any(fd == source for _, fd in timeout_queue))
    # Also expect at least one process-level timeout (fd=None)
    assert_true(any(fd is None for _, fd in timeout_queue))


def test_exit_3() -> None:
    # Expect the process to be closed after
    # the generator exits.
    rt = ThreadedRunner(cmd=["sleep", "4"],
                        stdin=None,
                        protocol_class=GenStdoutStderr,
                        timeout=.5,
                        exception_on_error=False)
    tuple(rt.run())
    assert_true(rt.return_code is not None)


def test_exit_4() -> None:
    rt = ThreadedRunner(cmd=["sleep", "4"],
                        stdin=None,
                        protocol_class=GenNothing,
                        timeout=.5)
    tuple(rt.run())
    assert_true(rt.return_code is not None)


def test_generator_throw() -> None:
    rt = ThreadedRunner(cmd=["sleep", "4"],
                        stdin=None,
                        protocol_class=GenNothing,
                        timeout=.5)
    gen = rt.run()
    assert isinstance(gen, Generator)
    assert_raises(ValueError, gen.throw, ValueError, ValueError("abcdefg"))


def test_exiting_process() -> None:
    result = run_command(py2cmd("import time\ntime.sleep(3)\nprint('exit')"),
                         protocol=NoCapture,
                         stdin=None)
    assert isinstance(result, dict)
    eq_(result["code"], 0)


def test_stalling_detection_1() -> None:
    runner = ThreadedRunner("something", StdOutErrCapture, None)
    runner.stdout_enqueueing_thread = None
    runner.stderr_enqueueing_thread = None
    runner.process_waiting_thread = None
    with patch("datalad.runner.nonasyncrunner.lgr") as logger:
        runner.process_queue()
    eq_(logger.method_calls[0][0], "warning")
    eq_(logger.method_calls[0][1][0], "ThreadedRunner.process_queue(): stall detected")


def test_stalling_detection_2() -> None:
    thread_mock = MagicMock()
    thread_mock.is_alive.return_value = False
    runner = ThreadedRunner("something", StdOutErrCapture, None)
    runner.stdout_enqueueing_thread = thread_mock
    runner.stderr_enqueueing_thread = thread_mock
    runner.process_waiting_thread = thread_mock
    with patch("datalad.runner.nonasyncrunner.lgr") as logger:
        runner.process_queue()
    eq_(logger.method_calls[0][0], "warning")
    eq_(logger.method_calls[0][1][0], "ThreadedRunner.process_queue(): stall detected")


def test_concurrent_waiting_run() -> None:
    from threading import Thread

    threaded_runner = ThreadedRunner(
        py2cmd("import time; time.sleep(1)"),
        protocol_class=NoCapture,
        stdin=None,
    )

    start = time.time()

    number_of_threads = 5
    caller_threads = []
    for c in range(number_of_threads):
        caller_thread = Thread(target=threaded_runner.run)
        caller_thread.start()
        caller_threads.append(caller_thread)

    while caller_threads:
        t = caller_threads.pop()
        t.join()

    # If the threads are serialized, the duration should at least
    # be one second per thread.
    duration = time.time() - start
    assert duration >= 1.0 * number_of_threads


def test_concurrent_generator_reading() -> None:
    number_of_lines = 40
    number_of_threads = 100
    output_queue: Queue[tuple[int, Optional[str]]] = Queue()

    threaded_runner = ThreadedRunner(
        py2cmd(f"for i in range({number_of_lines}): print(f'result#{{i}}')"),
        protocol_class=GenStdoutLines,
        stdin=None,
    )
    result_generator = threaded_runner.run()

    def thread_main(thread_number: int, result_generator: Iterator[str], output_queue: Queue[tuple[int, Optional[str]]]) -> None:
        while True:
            try:
                output = next(result_generator)
            except StopIteration:
                output_queue.put((thread_number, None))
                break
            output_queue.put((thread_number, output))

    caller_threads = []
    for c in range(number_of_threads):
        caller_thread = Thread(
            target=thread_main,
            args=(c, result_generator, output_queue)
        )
        caller_thread.start()
        caller_threads.append(caller_thread)

    while caller_threads:
        t = caller_threads.pop()
        t.join()

    collected_outputs = [
        output_tuple[1]
        for output_tuple in output_queue.queue
        if output_tuple[1] is not None
    ]
    assert len(collected_outputs) == number_of_lines
    assert collected_outputs == [
        f"result#{i}"
        for i in range(number_of_lines)
    ]


def test_same_thread_reenter_detection() -> None:
    threaded_runner = ThreadedRunner(
        py2cmd(f"print('hello')"),
        protocol_class=GenStdoutLines,
        stdin=None,
    )
    threaded_runner.run()
    with pytest.raises(RuntimeError) as error:
        threaded_runner.run()
    assert "re-entered by already" in str(error.value)


def test_reenter_generator_detection() -> None:
    threaded_runner = ThreadedRunner(
        py2cmd(f"print('hello')"),
        protocol_class=GenStdoutLines,
        stdin=None,
    )

    def target(threaded_runner: ThreadedRunner, output_queue: Queue[tuple[str, float | BaseException]]) -> None:
        try:
            start_time = time.time()
            tuple(threaded_runner.run())
            output_queue.put(("result", time.time() - start_time))
        except RuntimeError as exc:
            output_queue.put(("exception", exc))

    output_queue: Queue[tuple[str, float | BaseException]] = Queue()

    for sleep_time in range(1, 4):
        other_thread = Thread(
            target=target,
            args=(threaded_runner, output_queue)
        )

        gen = threaded_runner.run()
        other_thread.start()
        time.sleep(sleep_time)
        tuple(gen)
        other_thread.join()

        assert len(list(output_queue.queue)) == 1
        result_type, value = output_queue.get()
        assert result_type == "result"
        assert isinstance(value, float)
        assert value >= sleep_time



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for openfiles detection."""

import os
import subprocess
import sys
import time
from pathlib import Path
from unittest.mock import (
    MagicMock,
    patch,
)

import pytest

from datalad.tests.utils_pytest import (
    skip_if_no_psutil,
    with_tree,
)

from ..openfiles import (
    _is_write_mode,
    _lsof_get_write_files,
    get_files_open_for_writing,
)


def _hold_open(path, mode, seconds=30):
    """Spawn a subprocess that holds *path* open in *mode*."""
    proc = subprocess.Popen(
        [sys.executable, '-c',
         f'fh = open({str(path)!r}, {mode!r}); '
         f'import time; time.sleep({seconds})'])
    time.sleep(0.5)  # let the child actually open the file
    return proc


@skip_if_no_psutil
@pytest.mark.ai_generated
def test_get_files_open_for_writing(tmp_path):
    """Check various scenarios in a single tmp_path to avoid per-test overhead."""
    testfile = tmp_path / "testfile.txt"
    testfile.write_text("hello")

    # -- empty input --
    assert get_files_open_for_writing([]) == {}

    # -- nonexistent path --
    assert get_files_open_for_writing(['/no/such/file/anywhere']) == {}

    # -- file exists but not open --
    assert get_files_open_for_writing([str(testfile)]) == {}

    # -- our own process IS detected (no self-exclusion) --
    fh = open(testfile, 'w')
    try:
        result = get_files_open_for_writing([str(testfile)])
        assert str(testfile) in result, \
            "Our own process should be detected"
    finally:
        fh.close()

    # -- read-only open should NOT be detected --
    proc_r = _hold_open(testfile, 'r')
    try:
        assert get_files_open_for_writing(
            [str(testfile)]) == {}, \
            "Read-only open must not be detected"
    finally:
        proc_r.terminate()
        proc_r.wait()

    # -- write open IS detected --
    proc_w = _hold_open(testfile, 'w')
    try:
        result = get_files_open_for_writing(
            [str(testfile)])
        assert str(testfile) in result, \
            f"Write-open file not detected: {result}"
        assert any(o['pid'] == proc_w.pid for o in result[str(testfile)])
    finally:
        proc_w.terminate()
        proc_w.wait()

    # -- append open IS detected --
    proc_a = _hold_open(testfile, 'a')
    try:
        result = get_files_open_for_writing(
            [str(testfile)])
        assert str(testfile) in result, \
            f"Append-open file not detected: {result}"
    finally:
        proc_a.terminate()
        proc_a.wait()

    # -- r+ (read-write) IS detected --
    proc_rw = _hold_open(testfile, 'r+')
    try:
        result = get_files_open_for_writing(
            [str(testfile)])
        assert str(testfile) in result, \
            f"r+ open file not detected: {result}"
    finally:
        proc_rw.terminate()
        proc_rw.wait()

    # -- symlink resolved: querying via symlink still detects --
    link = tmp_path / "link.txt"
    link.symlink_to(testfile)
    proc_sym = _hold_open(testfile, 'w')
    try:
        result = get_files_open_for_writing(
            [str(link)])
        assert str(link) in result, \
            f"Symlink query should detect open file: {result}"
    finally:
        proc_sym.terminate()
        proc_sym.wait()


# -- Unit tests for helpers --------------------------------------------------


@pytest.mark.ai_generated
def test_is_write_mode():
    """Test _is_write_mode with various mode/flags combinations."""
    # mode string present
    f = MagicMock(mode='r', flags=None)
    assert _is_write_mode(f) is False
    f.mode = 'w'
    assert _is_write_mode(f) is True
    f.mode = 'a'
    assert _is_write_mode(f) is True
    f.mode = 'r+'
    assert _is_write_mode(f) is True
    f.mode = 'rb'
    assert _is_write_mode(f) is False

    # mode empty, flags present
    f = MagicMock(mode='', flags=os.O_RDONLY)
    assert _is_write_mode(f) is False
    f.flags = os.O_WRONLY
    assert _is_write_mode(f) is True
    f.flags = os.O_RDWR
    assert _is_write_mode(f) is True

    # neither mode nor flags
    f = MagicMock(spec=[])  # no attributes at all
    assert _is_write_mode(f) is None


@pytest.mark.ai_generated
def test_lsof_get_write_files():
    """Test lsof output parsing."""
    lsof_output = (
        "p1234\n"
        "fcwd\n"
        "ar\n"
        "n/some/read/file\n"
        "f3\n"
        "aw\n"
        "n/some/write/file\n"
        "f4\n"
        "au\n"
        "n/some/readwrite/file\n"
        "f5\n"
        "ar\n"
        "n/some/other/read\n"
    )
    with patch('subprocess.check_output', return_value=lsof_output):
        result = _lsof_get_write_files(1234)
    assert result == {'/some/write/file', '/some/readwrite/file'}


@pytest.mark.ai_generated
def test_lsof_get_write_files_unavailable():
    """Test lsof fallback when lsof is not installed."""
    with patch('subprocess.check_output', side_effect=FileNotFoundError):
        result = _lsof_get_write_files(1234)
    assert result is None


@pytest.mark.ai_generated
def test_lsof_get_write_files_timeout():
    """Test lsof fallback on timeout."""
    with patch('subprocess.check_output',
               side_effect=subprocess.TimeoutExpired('lsof', 10)):
        result = _lsof_get_write_files(1234)
    assert result is None


@skip_if_no_psutil
@pytest.mark.ai_generated
@with_tree({'testfile.txt': 'hello'})
def test_lsof_fallback_triggered(path=None):
    """When _is_write_mode returns None, lsof fallback should be used."""
    testfile = Path(path) / "testfile.txt"

    proc_w = _hold_open(testfile, 'w')
    try:
        resolved = str(testfile.resolve())
        lsof_output = f"p{proc_w.pid}\nf3\naw\nn{resolved}\n"
        with patch(
            'datalad.support.openfiles._is_write_mode', return_value=None
        ), patch(
            'subprocess.check_output', return_value=lsof_output
        ):
            result = get_files_open_for_writing(
                [str(testfile)])
        assert str(testfile) in result, \
            f"lsof fallback should detect write-open file: {result}"
    finally:
        proc_w.terminate()
        proc_w.wait()


# -- CLI tests ---------------------------------------------------------------


@skip_if_no_psutil
@pytest.mark.ai_generated
@with_tree({'quiet.txt': 'hello'})
def test_cli_no_open_files(path=None):
    """CLI reports no open files for a quiet directory."""
    testfile = Path(path) / 'quiet.txt'
    result = subprocess.run(
        [sys.executable, '-m', 'datalad.support.openfiles', str(testfile)],
        capture_output=True, text=True, timeout=30)
    assert result.returncode == 0
    assert 'No files open for writing' in result.stdout


@skip_if_no_psutil
@pytest.mark.ai_generated
@with_tree({'a.txt': 'a', 'b.txt': 'b'})
def test_cli_with_directory(path=None):
    """CLI expands directories to files."""
    result = subprocess.run(
        [sys.executable, '-m', 'datalad.support.openfiles', path],
        capture_output=True, text=True, timeout=30)
    assert result.returncode == 0
    assert 'No files open for writing' in result.stdout


@skip_if_no_psutil
@pytest.mark.ai_generated
@with_tree({'busy.txt': 'hello'})
def test_cli_detects_open_file(path=None):
    """CLI exits with 1 and reports files open for writing."""
    testfile = Path(path) / 'busy.txt'
    proc = _hold_open(testfile, 'w')
    try:
        result = subprocess.run(
            [sys.executable, '-m', 'datalad.support.openfiles',
             str(testfile)],
            capture_output=True, text=True, timeout=30)
        assert result.returncode == 1
        assert 'PIDs:' in result.stdout
    finally:
        proc.terminate()
        proc.wait()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Interface to add content, and save modifications to a dataset

"""

__docformat__ = 'restructuredtext'

import logging
from functools import partial
from pathlib import Path

import datalad.utils as ut
from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)
from datalad.interface.common_opts import (
    jobs_opt,
    recursion_flag,
    recursion_limit,
    save_message_opt,
)
from datalad.interface.utils import (
    discover_dataset_trace_to_targets,
    get_tree_roots,
)
from datalad.support.annexrepo import AnnexRepo
from datalad.support.constraints import (
    EnsureNone,
    EnsureStr,
)
from datalad.support.exceptions import CommandError
from datalad.support.parallel import (
    ProducerConsumerProgressLog,
    no_subds_in_futures,
)
from datalad.support.param import Parameter
from datalad.utils import ensure_list

from .diff import diff_dataset
from .status import Status

lgr = logging.getLogger('datalad.core.local.save')

_STATUS_SKIP_KEYS = frozenset((
    'path', 'parentds', 'refds', 'status', 'action', 'logger'))


def _status_props(result):
    """Extract status properties from a result dict, dropping metadata keys."""
    return {k: v for k, v in result.items() if k not in _STATUS_SKIP_KEYS}


def _inject_sub_info(paths_by_ds, since_map, _since_sub_info, ds_path):
    """Inject changed subdatasets that diff_dataset missed on adjusted branches.

    On adjusted branches, diff_dataset reports subdatasets as clean because
    the index submodule pointer records the adjusted SHA (which doesn't
    change when the sub gets new commits).  This function uses pre-command
    sub HEADs captured by run_command to:

    - Override since_map with the true pre-command SHAs
    - Add changed subs to paths_by_ds so save_ds processes them
    - Walk up from each changed sub to ds_path, ensuring every ancestor
      has since_map and paths_by_ds entries for merge propagation
    """
    for sub_path, sub_pre_sha in _since_sub_info.items():
        sub_repo = Dataset(sub_path).repo
        if not sub_repo:
            continue
        cur_sha = sub_repo.get_hexsha()
        if cur_sha == sub_pre_sha:
            continue

        since_map[sub_path] = sub_pre_sha
        if sub_path not in paths_by_ds:
            paths_by_ds[sub_path] = {}

        # Walk up to ds_path, registering each child as modified in
        # its parent and ensuring since_map entries exist at each level.
        # Termination: normally `parent_path == ds_path`.  Defensive
        # second termination on `parent_path == child_path` covers the
        # case where `sub_path` is not a descendant of `ds_path` (which
        # `_parse_sub_status` from `git submodule status --recursive`
        # never produces, but a future caller might) — without it,
        # `Path('/').parent == Path('/')` would loop forever.
        child_path = sub_path
        while True:
            parent_path = str(ut.Path(child_path).parent)
            if parent_path == child_path:
                lgr.debug("sub %s not under ds %s; aborting ancestor walk",
                          sub_path, ds_path)
                break
            child_repo = Dataset(child_path).repo
            child_sha = child_repo.get_hexsha() if child_repo else cur_sha
            child_pre = _since_sub_info.get(child_path, child_sha)

            pds_status = paths_by_ds.setdefault(parent_path, {})
            pds_status[ut.Path(child_path)] = dict(
                type='dataset', state='modified',
                prev_gitshasum=child_pre, gitshasum=child_sha)

            if parent_path not in since_map:
                since_map[parent_path] = _since_sub_info.get(
                    parent_path,
                    Dataset(parent_path).repo.get_hexsha())

            if parent_path == ds_path:
                break
            child_path = parent_path


def _create_merge_commit(repo, pre_hexsha, msg):
    """Create a merge commit wrapping command-created commits.

    The merge has two parents:
      parent1 = pre_hexsha (pre-command state, keeps first-parent history clean)
      parent2 = current HEAD (command's commits + any leftover save)
      tree    = current HEAD's tree (reflects all changes)

    On adjusted branches (git-annex), the merge is created on the
    original branch and then propagated back via ``git annex merge``.

    Raises
    ------
    CommandError
        If pre_hexsha is not an ancestor of the current HEAD, indicating
        the command switched branches or reset to an unrelated state.

    Parameters
    ----------
    repo : GitRepo or AnnexRepo
    pre_hexsha : str
        Commit hash of the pre-command state (first parent).
    msg : str
        Commit message (typically the run record).
    """
    on_adjusted = isinstance(repo, AnnexRepo) and repo.is_managed_branch()
    if on_adjusted:
        orig_branch = repo.get_corresponding_branch()
        repo.call_git([
            "annex", "sync", "--no-push", "--no-pull",
            "--no-commit", "--no-resolvemerge", "--no-content"])
        orig_post = repo.get_hexsha("refs/heads/" + orig_branch)
        orig_pre = repo.call_git_oneline(
            ["merge-base", orig_branch, pre_hexsha])
        # Verify the pre-command state is an ancestor of the
        # post-command state.  If not, the command switched branches.
        if not repo.call_git_success(
                ["merge-base", "--is-ancestor", orig_pre, orig_post]):
            raise CommandError(
                cmd="run",
                msg="cannot create merge commit: pre-command state "
                    f"{orig_pre[:8]} is not an ancestor of "
                    f"post-command state {orig_post[:8]} on branch "
                    f"{orig_branch}.  The command may have switched "
                    "branches or reset to an unrelated state.")
        tree = repo.call_git_oneline(
            ["rev-parse", orig_post + "^{tree}"])
        merge_hexsha = repo.call_git_oneline(
            ["commit-tree", tree,
             "-p", orig_pre, "-p", orig_post,
             "-m", msg])
        repo.update_ref("refs/heads/" + orig_branch, merge_hexsha)
        repo.call_git(["annex", "merge"])
    else:
        current_head = repo.get_hexsha()
        # Verify the pre-command state is an ancestor of the current
        # HEAD.  If not, the command switched branches or reset to an
        # unrelated state — creating a merge would be nonsensical.
        if not repo.call_git_success(
                ["merge-base", "--is-ancestor",
                 pre_hexsha, current_head]):
            raise CommandError(
                cmd="run",
                msg="cannot create merge commit: pre-command state "
                    f"{pre_hexsha[:8]} is not an ancestor of current "
                    f"HEAD {current_head[:8]}.  The command may have "
                    "switched branches or reset to an unrelated state.")
        tree = repo.call_git_oneline(
            ["rev-parse", current_head + "^{tree}"])
        merge_hexsha = repo.call_git_oneline(
            ["commit-tree", tree,
             "-p", pre_hexsha,
             "-p", current_head,
             "-m", msg])
        branch = repo.get_active_branch()
        if branch:
            repo.update_ref(
                "refs/heads/" + branch, merge_hexsha)
        else:
            repo.update_ref("HEAD", merge_hexsha)


@build_doc
class Save(Interface):
    """Save the current state of a dataset

    Saving the state of a dataset records changes that have been made to it.
    This change record is annotated with a user-provided description.
    Optionally, an additional tag, such as a version, can be assigned to the
    saved state. Such tag enables straightforward retrieval of past versions at
    a later point in time.

    .. note::
      Before Git v2.22, any Git repository without an initial commit located
      inside a Dataset is ignored, and content underneath it will be saved to
      the respective superdataset. DataLad datasets always have an initial
      commit, hence are not affected by this behavior.
    """
    # note above documents that out behavior is like that of `git add`, but
    # does not explicitly mention the connection to keep it simple.

    _examples_ = [
        dict(text="""Save any content underneath the current directory, without
             altering any potential subdataset""",
             code_py="save(path='.')",
             code_cmd="datalad save ."),
        dict(text="""Save specific content in the dataset""",
             code_py="save(path='myfile.txt')",
             code_cmd="datalad save myfile.txt"),
        dict(text="""Attach a commit message to save""",
             code_py="save(path='myfile.txt', message='add file')",
             code_cmd="datalad save -m 'add file' myfile.txt"),
        dict(text="""Save any content underneath the current directory, and
             recurse into any potential subdatasets""",
             code_py="save(path='.', recursive=True)",
             code_cmd="datalad save . -r"),
        dict(text="Save any modification of known dataset content in the "
                  "current directory, but leave untracked files (e.g. temporary files) "
                  "untouched",
             code_py="""save(path='.', updated=True)""",
             code_cmd="""datalad save -u ."""),
        dict(text="Tag the most recent saved state of a dataset",
             code_py="save(version_tag='bestyet')",
             code_cmd="datalad save --version-tag 'bestyet'"),
        dict(text="Save a specific change but integrate into last commit keeping "
                  "the already recorded commit message",
             code_py="save(path='myfile.txt', amend=True)",
             code_cmd="datalad save myfile.txt --amend")
    ]

    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc=""""specify the dataset to save""",
            constraints=EnsureDataset() | EnsureNone()),
        path=Parameter(
            args=("path",),
            metavar='PATH',
            doc="""path/name of the dataset component to save. If given, only
            changes made to those components are recorded in the new state.""",
            nargs='*',
            constraints=EnsureStr() | EnsureNone()),
        message=save_message_opt,
        message_file=Parameter(
            args=("-F", "--message-file"),
            doc="""take the commit message from this file. This flag is
            mutually exclusive with -m.""",
            constraints=EnsureStr() | EnsureNone()),
        version_tag=Parameter(
            args=("-t", "--version-tag",),
            metavar='ID',
            doc="""an additional marker for that state. Every dataset that
            is touched will receive the tag.""",
            constraints=EnsureStr() | EnsureNone()),
        recursive=recursion_flag,
        recursion_limit=recursion_limit,
        updated=Parameter(
            args=('-u', '--updated',),
            action='store_true',
            doc="""if given, only saves previously tracked paths."""),
        to_git=Parameter(
            args=("--to-git",),
            action='store_true',
            doc="""flag whether to add data directly to Git, instead of
            tracking data identity only.  Use with caution, there is no
            guarantee that a file put directly into Git like this will
            not be annexed in a subsequent save operation.
            If not specified, it will be up to git-annex to decide how
            a file is tracked, based on a dataset's configuration
            to track particular paths,
            file types, or file sizes with either Git or git-annex.
            (see https://git-annex.branchable.com/tips/largefiles).
            """),
        jobs=jobs_opt,
        amend=Parameter(
            args=('--amend',),
            action='store_true',
            doc="""if set, changes are not recorded in a new, separate
            commit, but are integrated with the changeset of the previous
            commit, and both together are recorded by replacing that
            previous commit. This is mutually exclusive with recursive
            operation.
            """),
        since=Parameter(
            args=("--since",),
            metavar='COMMIT',
            doc="""if given, compare against this commit instead of HEAD to
            discover changes, and wrap any intermediate commits made since
            that point as a merge commit. The merge has first-parent = since
            (keeping first-parent history linear) and second-parent = the
            post-save HEAD (all changes). This is used by `datalad run` to
            wrap command-created commits but can also be used standalone to
            close a unit of work as a merge.""",
            constraints=EnsureStr() | EnsureNone()),
    )

    @staticmethod
    @datasetmethod(name='save')
    @eval_results
    def __call__(path=None,
                 *,
                 message=None, dataset=None,
                 version_tag=None,
                 recursive=False, recursion_limit=None,
                 updated=False,
                 message_file=None,
                 to_git=None,
                 jobs=None,
                 amend=False,
                 since=None,
                 _since_sub_info=None,
                 _sub_message=None,
                 ):
        if message and message_file:
            raise ValueError(
                "Both a message and message file were specified for save()")

        if amend and recursive:
            raise ValueError("Cannot amend a commit recursively.")

        if amend and since is not None:
            raise ValueError(
                "Cannot amend when using --since; "
                "--since creates a merge commit.")

        path = ensure_list(path)

        if message_file:
            with open(message_file) as mfh:
                message = mfh.read()

        # we want 'normal' to achieve the most compact argument list
        # for git calls
        # untracked_mode = 'no' if updated else 'normal'
        # TODO however, Repo.add() would refuse to add any dotfiles
        # in a directory that is itself untracked, hence the only
        # choice is to go with potentially crazy long lists
        # until https://github.com/datalad/datalad/issues/1454
        # has a resolution
        untracked_mode = 'no' if updated else 'all'

        # there are three basic scenarios:
        # 1. save modifications to any already tracked content
        # 2. save any content (including removal of deleted content)
        #    to bring things to a clean state
        # 3. like (2), but only operate on a given subset of content
        #    identified by paths
        # - all three have to work in conjunction with --recursive
        # - the difference between (1) and (2) should be no more
        #   that a switch from --untracked=no to --untracked=all
        #   in Repo.save()

        # we do not support
        # - simultaneous operations on multiple datasets from disjoint
        #   dataset hierarchies, hence a single reference dataset must be
        #   identifiable from the either
        #   - curdir or
        #   - the `dataset` argument.
        #   This avoids complex annotation loops and hierarchy tracking.
        # - any modification upwards from the root dataset

        ds = require_dataset(dataset, check_installed=True, purpose='save')

        paths_by_ds = {}
        since_map = {}  # {ds_path: pre_hexsha} — only populated when since is set

        if since is not None:
            # Use diff_dataset to discover changes from the given baseline.
            # This translates `since` per-subdataset via prev_gitshasum.
            # ATTN: pass `dataset` (the original argument) rather than
            # `ds` (a Dataset instance) so that resolve_path() inside
            # diff_dataset resolves relative paths against CWD — the
            # same semantics as the Status() call in the else branch.
            # See resolve_path() docs: only Dataset *instances* trigger
            # dataset-root-relative resolution.  When dataset=None,
            # diff_dataset discovers the dataset from CWD via its own
            # require_dataset() call, same as Status does.
            for s in diff_dataset(
                    dataset=dataset, fr=since, to=None,
                    constant_refs=False,
                    # Stringify to avoid mixed PosixPath/str in
                    # diff_dataset's sorted() call
                    path=[str(p) for p in path] if path else None,
                    recursive=recursive,
                    recursion_limit=recursion_limit,
                    reporting_order='bottom-up',
                    # Use 'normal' (not untracked_mode='all') to
                    # report untracked directories as single entries
                    # rather than enumerating every file.  save_()
                    # handles directory entries correctly (git add
                    # recurses into them).  This matches the prior
                    # run.py behavior and avoids building a huge
                    # paths_by_ds dict for repos with large untracked
                    # trees (node_modules/, build/, etc).
                    untracked='normal' if not updated else 'no'):
                if s.get('status') == 'error':
                    yield s
                    continue
                if s.get('status') != 'ok':
                    continue
                ds_status = paths_by_ds.get(s['parentds'], {})
                props = _status_props(s)
                ds_status[ut.Path(s['path'])] = props
                paths_by_ds[s['parentds']] = ds_status
                # Track per-subdataset pre-command pointers for merge
                if (s.get('type') == 'dataset'
                        and 'prev_gitshasum' in props):
                    since_map[s['path']] = props['prev_gitshasum']
            # The top-level dataset's baseline is `since` itself
            since_map[ds.path] = since

            # On adjusted branches, diff_dataset may not detect
            # subdataset commits (index submodule pointer stays the
            # same).  Inject changed subs from run_command's snapshot.
            if _since_sub_info:
                _inject_sub_info(paths_by_ds, since_map,
                                 _since_sub_info, ds.path)
        else:
            # Standard Status-based discovery
            for s in Status()(
                    # ATTN: it is vital to pass the `dataset` argument as
                    # it, and not a dataset instance in order to maintain
                    # the path semantics between here and the status() call
                    dataset=dataset,
                    path=path,
                    untracked=untracked_mode,
                    recursive=recursive,
                    recursion_limit=recursion_limit,
                    on_failure='ignore',
                    eval_subdataset_state='full'
                    if recursive else 'commit',
                    return_type='generator',
                    result_renderer='disabled'):
                if s['status'] == 'error':
                    yield s
                    continue

                # fish out status dict for this parent dataset
                ds_status = paths_by_ds.get(s['parentds'], {})
                ds_status[ut.Path(s['path'])] = _status_props(s)
                paths_by_ds[s['parentds']] = ds_status

        lgr.debug('Determined %i datasets for saving from input arguments',
                  len(paths_by_ds))
        # figure out what datasets to process, start with the ones containing
        # the paths that were given as arguments
        discovered_datasets = list(paths_by_ds.keys())
        if dataset:
            # if a reference dataset was given we want to save all the way up
            # to it, so let's throw it into the mix
            discovered_datasets.append(ds.path)
        # sort the datasets into (potentially) disjoint hierarchies,
        # or a single one, if a reference dataset was given
        dataset_hierarchies = get_tree_roots(discovered_datasets)
        for rootds, children in dataset_hierarchies.items():
            edges = {}
            discover_dataset_trace_to_targets(
                rootds, children, [], edges, includeds=children)
            for superds, subdss in edges.items():
                superds_status = paths_by_ds.get(superds, {})
                for subds in subdss:
                    subds_path = ut.Path(subds)
                    sub_status = superds_status.get(subds_path, {})
                    if not (sub_status.get("state") == "clean" and
                            sub_status.get("type") == "dataset"):
                        # start from an entry that may already exist in the
                        # status record
                        superds_status[subds_path] = superds_status.get(
                            subds_path,
                            # if we got nothing yet:
                            # shot from the hip, some status config
                            # to trigger this specific super/sub
                            # relation to be saved
                            dict(state='untracked', type='dataset')
                        )
                paths_by_ds[superds] = superds_status

        def save_ds(args, version_tag=None):
            pdspath, paths = args

            pds = Dataset(pdspath)
            pds_repo = pds.repo
            # pop status for this dataset, we are not coming back to it
            pds_status = {
                # for handing over to the low-level code, we recode any
                # path relative to the real repo location, this avoid
                # cumbersome symlink handling without context in the
                # lower levels
                pds_repo.pathobj / p.relative_to(pdspath): props
                for p, props in paths.items()}
            start_commit = pds_repo.get_hexsha()

            # Determine whether this dataset needs a merge commit.
            # A merge is needed when either:
            #  - the command created intermediate commits (pre-command
            #    pointer != current HEAD), or
            #  - a child subdataset was merged (upward propagation).
            pre_hexsha = since_map.get(pdspath)
            had_inner = (pre_hexsha is not None
                         and pre_hexsha != start_commit)
            child_merged = (
                pre_hexsha is not None
                and any(str(p) in _merged_datasets
                        for p, props in paths.items()
                        if props.get('type') == 'dataset'))
            will_merge = had_inner or child_merged

            if not all(p['state'] == 'clean'
                       for p in pds_status.values()) or \
                    (amend and message):
                # When wrapping inner commits as a merge, the intermediate
                # save of working-tree changes is subordinate to the merge
                # commit (it lives on the second-parent chain).  Use
                # _sub_message if the caller provided one (run_command
                # passes a run-specific string); otherwise fall back to the
                # user's message.
                save_msg = (_sub_message if will_merge and _sub_message
                            else message)
                for res in pds_repo.save_(
                        message=save_msg,
                        # make sure to have the `path` arg be None, as we
                        # want to prevent and bypass any additional
                        # repo.status() calls
                        paths=None,
                        # prevent whining of GitRepo
                        git=True if not hasattr(pds_repo, 'uuid')
                        else to_git,
                        # we are supplying the full status already, do not
                        # detect anything else
                        untracked='no',
                        _status=pds_status,
                        amend=amend):
                    for k in ('path', 'refds'):
                        if k in res:
                            res[k] = str(
                                # recode path back to dataset path anchor
                                pds.pathobj / Path(res[k]).relative_to(
                                    pds_repo.pathobj)
                            )
                    yield res

            if will_merge:
                try:
                    _create_merge_commit(pds_repo, pre_hexsha, message)
                except CommandError as exc:
                    yield dict(
                        action='save', type='dataset',
                        path=pds.path, refds=ds.path,
                        status='error',
                        message=str(exc),
                        logger=lgr)
                    return
                _merged_datasets.add(pdspath)

            # report on the dataset itself
            dsres = dict(
                action='save',
                type='dataset',
                path=pds.path,
                refds=ds.path,
                status='ok'
                if start_commit != pds_repo.get_hexsha()
                else 'notneeded',
                logger=lgr,
            )
            if not version_tag:
                yield dsres
                return
            try:
                # method requires str
                version_tag = str(version_tag)
                pds_repo.tag(version_tag)
                dsres.update(
                    status='ok',
                    version_tag=version_tag)
                yield dsres
            except CommandError as e:
                if dsres['status'] == 'ok':
                    # first we yield the result for the actual save
                    yield dsres.copy()
                # and now complain that tagging didn't work
                dsres.update(
                    status='error',
                    message=('cannot tag this version: %s', e.stderr.strip()))
                yield dsres

        # Track which datasets got merge commits, so parent save_ds
        # can detect child merges for upward propagation.
        # Safe because ProducerConsumerProgressLog with
        # no_subds_in_futures processes children before parents.
        _merged_datasets = set()

        if not paths_by_ds:
            # Special case: empty repo. There's either an empty commit only or
            # none at all. An empty one we can amend otherwise there's nothing
            # to do.
            if amend and ds.repo.get_hexsha():
                yield from save_ds((ds.pathobj, dict()), version_tag=version_tag)

            else:
                yield dict(action='save',
                           type='dataset',
                           path=ds.path,
                           refds=ds.path,
                           status='notneeded',
                           logger=lgr)
            return

        # TODO: in principle logging could be improved to go not by a dataset
        # but by path(s) within subdatasets. That should provide a bit better ETA
        # and more "dynamic" feedback than jumpy datasets count.
        # See addurls where it is implemented that way by providing agg and another
        # log_filter
        yield from ProducerConsumerProgressLog(
            sorted(paths_by_ds.items(), key=lambda v: v[0], reverse=True),
            partial(save_ds, version_tag=version_tag),
            safe_to_consume=no_subds_in_futures,
            producer_future_key=lambda ds_items: ds_items[0],
            jobs=jobs,
            log_filter=_log_filter_save_dataset,
            unit="datasets",
            lgr=lgr,
        )


def _log_filter_save_dataset(res):
    return res.get('type') == 'dataset' and res.get('action') == 'save'



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Detect files open for writing by other processes.

Provides :func:`get_files_open_for_writing` which checks whether any of the
given paths are held open for writing by another process.  Uses *psutil*.

Can also be invoked as ``python -m datalad.support.openfiles``.
"""

from __future__ import annotations

import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Any

from datalad.utils import on_windows

try:
    import psutil
except ImportError:
    psutil = None  # type: ignore[assignment]

lgr = logging.getLogger('datalad.support.openfiles')

# ---------------------------------------------------------------------------
# Public API
# ---------------------------------------------------------------------------

def get_files_open_for_writing(
    paths: list[str | Path],
) -> dict[str, list[dict[str, Any]]]:
    """Return a mapping of *paths* that are open for writing.

    Parameters
    ----------
    paths
        File paths to check (absolute or relative – they will be resolved).

    Returns
    -------
    dict
        ``{original_path: [{'pid': int, 'fd': int}, ...]}`` for every path
        from *paths* that is currently open for writing by at least one
        process.

    Raises
    ------
    ImportError
        If *psutil* is not installed.
    """
    if psutil is None:
        raise ImportError(
            "psutil is required for open-file detection; "
            "install it with: pip install datalad[misc]")

    if not paths:
        return {}

    # Build a lookup: resolved_path -> original_path
    resolved_to_orig: dict[str, str] = {}
    for p in paths:
        try:
            resolved = str(Path(p).resolve())
        except OSError:
            lgr.log(5, "Cannot resolve %r, skipping", p)
            continue
        resolved_to_orig[resolved] = str(p)

    if not resolved_to_orig:
        return {}

    lgr.log(5, "Checking %d path(s) for open-for-writing fds", len(resolved_to_orig))

    target_paths = set(resolved_to_orig)
    result: dict[str, list[dict[str, Any]]] = {}

    # Only inspect processes owned by the current user — other users'
    # processes would raise AccessDenied anyway and cannot conflict
    # with our git operations.
    # On Windows os.getuid() and psutil's 'uids' attr are unavailable,
    # so UID-based filtering is skipped and all processes are scanned.
    _have_uids = hasattr(os, 'getuid')
    my_uid = os.getuid() if _have_uids else None
    n_procs = 0
    n_skipped_uid = 0
    # Cache lsof results per-pid so we call it at most once per process
    _lsof_cache: dict[int, set[str] | None] = {}
    # Cache Windows CreateFileW probe results per-path
    _win_write_cache: dict[str, bool | None] = {}

    _iter_attrs = ['pid'] + (['uids'] if _have_uids else [])
    for proc in psutil.process_iter(_iter_attrs):
        try:
            pid = proc.info['pid']
            proc_uids = proc.info.get('uids') if _have_uids else None
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            continue
        # Skip processes owned by other users (not available on Windows)
        if my_uid is not None and proc_uids is not None and proc_uids.real != my_uid:
            n_skipped_uid += 1
            continue
        n_procs += 1
        try:
            open_files = proc.open_files()
        except (psutil.NoSuchProcess, psutil.AccessDenied) as exc:
            if lgr.isEnabledFor(logging.DEBUG):
                _proc_info = _describe_process(proc)
                lgr.debug("Cannot read open files of %s: %s", _proc_info, exc)
            continue
        for f in open_files:
            if f.path not in target_paths:
                continue
            if lgr.isEnabledFor(logging.DEBUG):
                _proc_info = _describe_process(proc)
                lgr.debug("Target file %s open by %s mode=%r fd=%s",
                          f.path, _proc_info,
                          getattr(f, 'mode', ''),
                          getattr(f, 'fd', '?'))
            is_write = _is_write_mode(f)
            if is_write is None:
                # mode/flags unavailable (e.g. macOS) — fall back to lsof
                if pid not in _lsof_cache:
                    _lsof_cache[pid] = _lsof_get_write_files(pid)
                lsof_writes = _lsof_cache[pid]
                if lsof_writes is not None:
                    is_write = f.path in lsof_writes
            if is_write is None and on_windows:
                # Windows: no mode/flags or lsof; probe via CreateFileW
                if f.path not in _win_write_cache:
                    _win_write_cache[f.path] = _win_is_open_for_writing(
                        f.path)
                is_write = _win_write_cache[f.path]
            if is_write is None:
                lgr.log(5, "Cannot determine open mode for %s "
                           "(pid=%d), skipping", f.path, pid)
                continue
            if is_write:
                orig = resolved_to_orig.get(f.path, f.path)
                result.setdefault(orig, []).append(
                    {'pid': pid,
                     'fd': getattr(f, 'fd', -1)})

    lgr.log(5, "Scanned %d procs (skipped %d other-user), "
               "found %d path(s) open for writing",
            n_procs, n_skipped_uid, len(result))

    return result


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _is_write_mode(f) -> bool | None:
    """Check whether open-file handle *f* is opened for writing.

    Returns ``True``/``False`` when determinable, ``None`` when neither
    *mode* nor *flags* are available (e.g. macOS with older psutil).
    """
    mode = getattr(f, 'mode', '')
    if mode:
        return any(c in mode for c in ('w', 'a', '+'))
    flags = getattr(f, 'flags', None)
    if flags is not None:
        return bool(flags & (os.O_WRONLY | os.O_RDWR))
    return None


def _lsof_get_write_files(pid: int) -> set[str] | None:
    """Return resolved paths open for writing by *pid*, via ``lsof``.

    Falls back to ``lsof`` on platforms where *psutil* does not expose
    the file-open mode (notably macOS).  Returns ``None`` when ``lsof``
    is unavailable or fails.
    """
    try:
        out = subprocess.check_output(
            ['lsof', '-p', str(pid), '-F', 'afn'],
            text=True, timeout=10, stderr=subprocess.DEVNULL)
    except (subprocess.SubprocessError, FileNotFoundError, OSError) as exc:
        lgr.log(5, "lsof fallback failed for pid %d: %s", pid, exc)
        return None

    write_files: set[str] = set()
    current_access: str | None = None
    for line in out.splitlines():
        if line.startswith('a'):
            current_access = line[1:]  # 'r', 'w', or 'u' (read-write)
        elif line.startswith('n') and current_access in ('w', 'u'):
            write_files.add(line[1:])
    return write_files


def _win_is_open_for_writing(path: str) -> bool | None:
    """Check whether *path* is held open for writing, on Windows.

    Probes with ``CreateFileW`` requesting write access with read-only
    sharing.  A ``ERROR_SHARING_VIOLATION`` (32) indicates another
    process holds a write handle on the file.

    Returns ``True`` (write-open detected), ``False`` (no write
    conflict), or ``None`` (cannot determine).
    """
    try:
        import ctypes
        from ctypes import wintypes
    except ImportError:
        return None

    kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
    kernel32.CreateFileW.restype = wintypes.HANDLE
    kernel32.CloseHandle.argtypes = [wintypes.HANDLE]

    _GENERIC_WRITE = 0x40000000
    _FILE_SHARE_READ = 0x00000001
    _OPEN_EXISTING = 3

    handle = kernel32.CreateFileW(
        path, _GENERIC_WRITE, _FILE_SHARE_READ,
        None, _OPEN_EXISTING, 0, None)
    # INVALID_HANDLE_VALUE = (HANDLE)(-1)
    if handle == ctypes.c_void_p(-1).value or handle is None:
        err = ctypes.get_last_error()
        if err == 32:  # ERROR_SHARING_VIOLATION
            return True
        return None
    kernel32.CloseHandle(handle)
    return False


def _describe_process(proc: psutil.Process) -> str:
    """Return a short description of *proc* for log messages."""
    try:
        cmdline = proc.cmdline()
        cmd = ' '.join(cmdline) if cmdline else '?'
    except (psutil.NoSuchProcess, psutil.AccessDenied):
        cmd = '?'
    try:
        username = proc.username()
    except (psutil.NoSuchProcess, psutil.AccessDenied):
        username = '?'
    return f"pid {proc.pid} ({cmd!r}, user={username})"



# ---------------------------------------------------------------------------
# CLI
# ---------------------------------------------------------------------------

def _cli_main() -> None:
    import argparse

    parser = argparse.ArgumentParser(
        description='Show files open for writing under given paths.')
    parser.add_argument('paths', nargs='*', default=['.'],
                        help='Files or directories to check.  Directories '
                             'are automatically expanded to all files '
                             'underneath.')
    args = parser.parse_args()

    paths: list[str] = []
    for p in args.paths:
        if os.path.isdir(p):
            # TODO: consider skipping .git/ and other VCS dirs, or
            # limiting to files that would be subject to commit
            for root, _dirs, files in os.walk(p):
                paths.extend(os.path.join(root, f) for f in files)
        else:
            paths.append(p)

    result = get_files_open_for_writing(paths)
    if result:
        for path, openers in sorted(result.items()):
            pids = ', '.join(str(o['pid']) for o in openers)
            print(f'{path}  (PIDs: {pids})')
        sys.exit(1)
    else:
        print('No files open for writing.')


if __name__ == '__main__':
    _cli_main()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""A Helper to initiate arbitrarily small/large test meta-dataset

"""

__docformat__ = 'numpy'


import logging
import os
import random
import tempfile
from os.path import (
    abspath,
    exists,
    isabs,
)
from os.path import join as opj

from datalad.distribution.dataset import Dataset
from datalad.interface.base import (
    Interface,
    build_doc,
)
from datalad.support.annexrepo import AnnexRepo
from datalad.support.constraints import (
    EnsureInt,
    EnsureNone,
    EnsureStr,
)
from datalad.support.gitrepo import GitRepo
from datalad.support.param import Parameter
from datalad.utils import get_tempfile_kwargs

lgr = logging.getLogger('datalad.distribution.tests')


def _parse_spec(spec):
    """Parse a hierarchy spec string into a list of level descriptors.

    Each level is separated by ``/`` and has the form::

        [d]min[-max][+Nf]

    ``d`` prefix makes entries plain directories instead of subdatasets.
    ``+Nf`` suffix adds N tracked files to each dataset at that level
    (default: 1 file per dataset, as before).

    Returns a list of dicts with keys: min, max, is_dir, nfiles.
    """
    out = []
    if not spec:
        return out
    for ilevel, level in enumerate(spec.split('/')):
        if not level:
            continue

        is_dir = False
        nfiles = None  # None means "use the nfiles parameter default"

        # Check for +Nf suffix (e.g., "2+100f")
        if '+' in level and level.endswith('f'):
            level, nf_str = level.rsplit('+', 1)
            nfiles = int(nf_str[:-1])  # strip trailing 'f'

        # Check for 'd' prefix (plain directory, not subdataset)
        if level.startswith('d'):
            is_dir = True
            level = level[1:]

        minmax = level.split('-')
        if len(minmax) == 1:  # only abs number specified
            minmax = int(minmax[0])
            min_, max_ = (minmax, minmax)
        elif len(minmax) == 2:
            min_, max_ = minmax
            if not min_:  # might be omitted entirely
                min_ = 0
            if not max_:
                raise ValueError(
                    "Specify max number at level %d. Full spec was: %s"
                    % (ilevel, spec))
            min_ = int(min_)
            max_ = int(max_)
        else:
            raise ValueError(
                "Must have only min-max at level %d" % ilevel)

        out.append(dict(min=min_, max=max_, is_dir=is_dir, nfiles=nfiles))
    return out


def _makeds(path, levels, ds=None, max_leading_dirs=2, nfiles=1):
    """Create a hierarchy of datasets

    Used recursively, with current invocation generating datasets for the
    first level, and delegating sub-levels to recursive invocation

    Parameters
    ----------
    path : str
      Path to the top directory under which dataset will be created.
      If relative -- relative to current directory
    levels : list of dict
      List of level descriptors from :func:`_parse_spec`.
    ds : Dataset, optional
      Super-dataset which would contain a new dataset (thus its path would be
      a parent of path. Note that ds needs to be installed.
    max_leading_dirs : int, optional
      Up to how many leading directories within a dataset could lead to a
      sub-dataset
    nfiles : int, optional
      Number of files to create in each dataset (default 1).  Per-level
      ``+Nf`` spec overrides this.

    Yields
    ------
    str
       Path to the generated dataset(s)

    """
    # we apparently can't import api functionality within api
    from datalad.api import save

    # To simplify managing all the file paths etc
    if not isabs(path):
        path = abspath(path)
    # make it a git (or annex??) repository... ok - let's do randomly one or another ;)
    RepoClass = GitRepo if random.randint(0, 1) else AnnexRepo
    lgr.info("Generating repo of class %s under %s", RepoClass, path)
    repo = RepoClass(path, create=True)
    # Create tracked files.  Consume one random.randint for the first
    # filename to preserve RNG sequence compatibility with the original
    # code that used random filenames (callers with fixed seeds depend
    # on the RNG producing the same dataset structure).
    fns = []
    first_file_id = random.randint(1, 1000)
    for fi in range(nfiles):
        fn = opj(path, "file%d.dat" % (first_file_id if fi == 0 else fi))
        with open(fn, 'w') as f:
            f.write(fn)
        fns.append(fn)
    repo.add(fns, git=True)
    repo.commit(msg="Added %d file(s)" % nfiles)

    yield path

    if levels:
        # make a dataset for that one since we want to add sub datasets
        ds_ = Dataset(path)
        # Process the levels
        level, levels_ = levels[0], levels[1:]
        nchildren = random.randint(level['min'], level['max'])
        child_nfiles = level.get('nfiles') or nfiles

        for ichild in range(nchildren):
            if level['is_dir']:
                # Plain directory with files, not a subdataset
                dirname = "dir%d" % ichild
                dirpath = opj(path, dirname)
                os.makedirs(dirpath, exist_ok=True)
                for fi in range(child_nfiles):
                    fn = opj(dirpath, "file%d.dat" % fi)
                    with open(fn, 'w') as f:
                        f.write(fn)
                repo.add(dirname, git=True)
            else:
                # Subdataset
                # we would like to have up to 2 leading dirs
                subds_path = opj(
                    *(['d%i' % i
                       for i in range(
                           random.randint(0, max_leading_dirs + 1))]
                      + ['r%i' % ichild]))
                subds_fpath = opj(path, subds_path)
                # yield all under
                for d in _makeds(subds_fpath, levels_, ds=ds_,
                                 nfiles=child_nfiles):
                    yield d

        if level['is_dir']:
            # Commit the directories we just created
            repo.commit(msg="Added %d directories" % nchildren)

    if ds:
        assert ds.is_installed()
        out = save(
            path,
            dataset=ds,
        )


@build_doc
class CreateTestDataset(Interface):
    """Create test (meta-)dataset.
    """

    _params_ = dict(
        path=Parameter(
            args=("path",),
            doc="path/name where to create (if specified, must not exist)",
            constraints=EnsureStr() | EnsureNone()),
        spec=Parameter(
            args=("--spec",),
            doc="""\
            spec for hierarchy.  Each level is separated by ``/`` and has
            the form ``[d]min[-max][+Nf]``.

            ``min-max`` defines how many (random from min to max)
            sub-datasets to generate at that level (min can be omitted
            to assume 0).

            Prefix ``d`` makes entries plain directories instead of
            subdatasets (e.g. ``d3`` creates 3 dirs with files).

            Suffix ``+Nf`` creates N tracked files per entry at that
            level (overrides --nfiles for that level).

            Examples::

                1-3/-2        1–3 subdatasets at L1, up to 2 at L2
                2+100f/-2     2 subs at L1 each with 100 files, up to 2 at L2
                10/d5+50f     10 subs, each containing 5 plain dirs with 50 files
            """,
            constraints=EnsureStr() | EnsureNone()),
        seed=Parameter(
            args=("--seed",),
            doc="""seed for rng""",
            constraints=EnsureInt() | EnsureNone()),
        nfiles=Parameter(
            args=("--nfiles",),
            doc="""number of files to create in each dataset (default 1).
            Per-level +Nf suffix in spec overrides this.""",
            constraints=EnsureInt() | EnsureNone()),
    )

    @staticmethod
    def __call__(path=None, *, spec=None, seed=None, nfiles=None):
        levels = _parse_spec(spec)

        if seed is not None:
            # TODO: if to be used within a bigger project we shouldn't seed main RNG
            random.seed(seed)
        if nfiles is None:
            nfiles = 1
        if path is None:
            kw = get_tempfile_kwargs({}, prefix="ds")
            path = tempfile.mkdtemp(**kw)
        else:
            # so we don't override anything
            assert not exists(path)
            os.makedirs(path)

        # now we should just make it happen and return list of all the datasets
        return list(_makeds(path, levels, nfiles=nfiles))

    @staticmethod
    def result_renderer_cmdline(res, args):
        from datalad.ui import ui
        if res is None:
            res = []
        if not len(res):
            ui.message("No repos were created... oops")
            return
        items = '\n'.join(map(str, res))
        msg = "{n} installed {obj} available at\n{items}".format(
            obj='items are' if len(res) > 1 else 'item is',
            n=len(res),
            items=items)
        ui.message(msg)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for DataLad pytest plugin.

These tests verify that the pytest plugin correctly registers markers,
configuration, and works as expected for DataLad extensions.

The tests use pytest's pytester fixture to run pytest in a subprocess,
avoiding double-registration issues when testing the plugin itself.
"""

import pytest

# Enable pytester fixture for plugin testing
pytest_plugins = ["pytester"]


@pytest.mark.ai_generated
def test_plugin_registers_markers(pytester):
    """Test that plugin registers custom markers."""
    pytester.makepyfile("""
        import pytest

        @pytest.mark.network
        def test_network():
            pass

        @pytest.mark.slow
        def test_slow():
            pass

        @pytest.mark.integration
        def test_integration():
            pass
    """)

    result = pytester.runpytest_subprocess("--markers")
    # Check that key markers are registered (not order-dependent)
    output = result.stdout.str()
    assert "@pytest.mark.network: marks tests requiring network access" in output
    assert "@pytest.mark.slow: marks slow tests" in output
    assert "@pytest.mark.integration: marks integration tests" in output


@pytest.mark.ai_generated
def test_plugin_markers_work(pytester):
    """Test that markers can be used for test selection."""
    pytester.makepyfile("""
        import pytest

        @pytest.mark.network
        def test_with_network():
            assert True

        def test_without_marker():
            assert True
    """)

    # Select only network tests
    result = pytester.runpytest_subprocess("-v", "-m", "network")
    result.stdout.fnmatch_lines(["*test_with_network*PASSED*"])
    assert "test_without_marker" not in result.stdout.str()


@pytest.mark.ai_generated
def test_plugin_python_files_pattern(pytester):
    """Test that utils_pytest.py is collected."""
    pytester.makepyfile(utils_pytest="""
        def test_in_utils_pytest():
            assert True
    """)

    # Run with collect-only to see if file is discovered
    result = pytester.runpytest_subprocess("-v", "--collect-only")
    # Check that utils_pytest.py is in the collected items
    assert "utils_pytest.py" in result.stdout.str(), \
        f"utils_pytest.py not collected. Output:\n{result.stdout.str()}"

    # Now run the actual test
    result = pytester.runpytest_subprocess("-v")
    result.stdout.fnmatch_lines(["*utils_pytest*test_in_utils_pytest*PASSED*"])


@pytest.mark.ai_generated
def test_plugin_ignores_nose_tests(pytester):
    """Test that old nose test files are ignored."""
    pytester.makepyfile(test_tests_utils="""
        def test_should_not_run():
            assert False, "This should be ignored"
    """)

    result = pytester.runpytest_subprocess("-v", "--collect-only")
    # test_tests_utils.py should be ignored
    assert "test_tests_utils" not in result.stdout.str()


@pytest.mark.ai_generated
def test_plugin_filterwarnings_configured(pytester):
    """Test that filterwarnings are registered."""
    # Create a test that would trigger warnings
    pytester.makepyfile("""
        import warnings

        def test_warning():
            # This test just checks that pytest can run
            # The actual filterwarnings behavior is harder to test
            # in isolation, but we verify the config is present
            assert True
    """)

    result = pytester.runpytest_subprocess("--help")
    # Just verify pytest runs successfully with our plugin
    assert result.ret == 0


@pytest.mark.ai_generated
def test_plugin_works_for_extensions(pytester):
    """Test simulating how an extension would use the plugin."""
    # Create an extension-like test structure
    pytester.makepyfile("""
        import pytest

        @pytest.mark.slow
        @pytest.mark.network
        def test_extension_functionality():
            '''Example extension test using datalad markers.'''
            assert True

        @pytest.mark.turtle
        def test_very_slow():
            '''Example of very slow test.'''
            assert True
    """)

    # Test that marker selection works
    result = pytester.runpytest_subprocess("-v", "-m", "slow and network")
    result.stdout.fnmatch_lines(["*test_extension_functionality*PASSED*"])
    assert "test_very_slow" not in result.stdout.str()

    # Test that turtle marker works
    result = pytester.runpytest_subprocess("-v", "-m", "turtle")
    result.stdout.fnmatch_lines(["*test_very_slow*PASSED*"])
    assert "test_extension_functionality" not in result.stdout.str()


@pytest.mark.ai_generated
def test_all_expected_markers_present(pytester):
    """Verify all expected markers are registered."""
    result = pytester.runpytest_subprocess("--markers")

    # Check for key markers
    expected_markers = [
        "network",
        "slow",
        "turtle",
        "integration",
        "known_failure",
        "skip_if_no_network",
        "windows",
        "osx",
        "ai_generated",
    ]

    for marker in expected_markers:
        assert marker in result.stdout.str(), f"Marker '{marker}' not found"


@pytest.mark.ai_generated
@pytest.mark.slow
def test_plugin_no_double_registration(pytester):
    """Test that the plugin doesn't cause double registration.

    This is a critical test ensuring the architectural fix works.
    When testing datalad itself, the hooks from pytest_plugin are
    imported into conftest, not registered via entry point, so there
    should be no double registration.
    """
    pytester.makepyfile("""
        import pytest

        @pytest.mark.network
        def test_example():
            assert True
    """)

    result = pytester.runpytest_subprocess("-v")
    # Should not have any plugin registration errors
    assert "Plugin already registered" not in result.stdout.str()
    assert "Plugin already registered" not in result.stderr.str()
    assert result.ret == 0
